# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import threading
import octoprint.vendor.sockjs.tornado
import octoprint.vendor.sockjs.tornado.session
import octoprint.vendor.sockjs.tornado.proto
import octoprint.vendor.sockjs.tornado.util
import time

import octoprint.timelapse
import octoprint.server
import octoprint.events
import octoprint.plugin
import octoprint.access.users

from octoprint.events import Events
from octoprint.settings import settings
from octoprint.access.permissions import Permissions
from octoprint.access.users import LoginStatusListener
from octoprint.access.groups import GroupChangeListener
from octoprint.util.json import JsonEncoding

import octoprint.printer

import wrapt
import json


class ThreadSafeSession(octoprint.vendor.sockjs.tornado.session.Session):
	def __init__(self, conn, server, session_id, expiry=None):
		octoprint.vendor.sockjs.tornado.session.Session.__init__(self, conn, server, session_id, expiry=expiry)

	def set_handler(self, handler, start_heartbeat=True):
		if getattr(handler, "__orig_send_pack", None) is None:
			orig_send_pack = handler.send_pack
			mutex = threading.RLock()

			def send_pack(*args, **kwargs):
				with mutex:
					return orig_send_pack(*args, **kwargs)

			handler.send_pack = send_pack
			setattr(handler, "__orig_send_pack", orig_send_pack)

		return octoprint.vendor.sockjs.tornado.session.Session.set_handler(self, handler, start_heartbeat=start_heartbeat)

	def remove_handler(self, handler):
		result = octoprint.vendor.sockjs.tornado.session.Session.remove_handler(self, handler)

		if getattr(handler, "__orig_send_pack", None) is not None:
			handler.send_pack = getattr(handler, "__orig_send_pack")
			delattr(handler, "__orig_send_pack")

		return result


class JsonEncodingSessionWrapper(wrapt.ObjectProxy):
	def send_message(self, msg, stats=True, binary=False):
		self.send_jsonified(json.dumps(octoprint.vendor.sockjs.tornado.util.bytes_to_str(msg),
		                               separators=(',', ':'),
		                               default=JsonEncoding.encode),
		                    stats)


class PrinterStateConnection(octoprint.vendor.sockjs.tornado.SockJSConnection,
                             octoprint.printer.PrinterCallback,
                             LoginStatusListener,
                             GroupChangeListener):

	_event_permissions = {Events.USER_LOGGED_IN: [Permissions.ADMIN],
	                      Events.USER_LOGGED_OUT: [Permissions.ADMIN],
	                      "*": []}

	_event_payload_processors = {Events.CLIENT_OPENED: [lambda user, payload: payload if user.has_permission(Permissions.ADMIN) else dict()],
	                             Events.CLIENT_AUTHED: [lambda user, payload: payload if user.has_permission(Permissions.ADMIN) else dict()],
	                             "*": []}

	_emit_permissions = {"connected": [],
	                     "reauthRequired": [],
	                     "*": [Permissions.STATUS]}

	_unauthed_backlog_max = 100

	def __init__(self, printer, fileManager, analysisQueue, userManager, groupManager, eventManager, pluginManager, session):
		if isinstance(session, octoprint.vendor.sockjs.tornado.session.Session):
			session = JsonEncodingSessionWrapper(session)

		octoprint.vendor.sockjs.tornado.SockJSConnection.__init__(self, session)

		self._logger = logging.getLogger(__name__)

		self._temperatureBacklog = []
		self._temperatureBacklogMutex = threading.Lock()
		self._logBacklog = []
		self._logBacklogMutex = threading.Lock()
		self._messageBacklog = []
		self._messageBacklogMutex = threading.Lock()

		self._unauthed_backlog = []
		self._unauthed_backlog_mutex = threading.RLock()

		self._printer = printer
		self._fileManager = fileManager
		self._analysisQueue = analysisQueue
		self._userManager = userManager
		self._groupManager = groupManager
		self._eventManager = eventManager
		self._pluginManager = pluginManager

		self._remoteAddress = None
		self._user = self._userManager.anonymous_user_factory()

		self._throttleFactor = 1
		self._lastCurrent = 0
		self._baseRateLimit = 0.5

		self._register_hooks = self._pluginManager.get_hooks("octoprint.server.sockjs.register")
		self._authed_hooks = self._pluginManager.get_hooks("octoprint.server.sockjs.authed")
		self._emit_hooks = self._pluginManager.get_hooks("octoprint.server.sockjs.emit")

		self._registered = False
		self._authed = False

	@staticmethod
	def _get_remote_address(info):
		forwarded_for = info.headers.get("X-Forwarded-For")
		if forwarded_for is not None:
			return forwarded_for.split(",")[0]
		return info.ip

	def __str__(self):
		if self._remoteAddress:
			return "{!r} connected to {}".format(self, self._remoteAddress)
		else:
			return "Unconnected {!r}".format(self)

	def on_open(self, info):
		self._pluginManager.register_message_receiver(self.on_plugin_message)
		self._remoteAddress = self._get_remote_address(info)
		self._logger.info("New connection from client: %s" % self._remoteAddress)

		self._userManager.register_login_status_listener(self)
		self._groupManager.register_listener(self)

		plugin_signature = lambda impl: "{}:{}".format(impl._identifier, impl._plugin_version)
		template_plugins = list(map(plugin_signature, self._pluginManager.get_implementations(octoprint.plugin.TemplatePlugin)))
		asset_plugins = list(map(plugin_signature, self._pluginManager.get_implementations(octoprint.plugin.AssetPlugin)))
		ui_plugins = sorted(set(template_plugins + asset_plugins))

		import hashlib
		plugin_hash = hashlib.md5()
		plugin_hash.update(",".join(ui_plugins).encode('utf-8'))

		config_hash = settings().config_hash

		# connected => update the API key, might be necessary if the client was left open while the server restarted
		self._emit("connected", dict(version=octoprint.server.VERSION,
		                             display_version=octoprint.server.DISPLAY_VERSION,
		                             branch=octoprint.server.BRANCH,
		                             plugin_hash=plugin_hash.hexdigest(),
		                             config_hash=config_hash,
		                             debug=octoprint.server.debug,
		                             safe_mode=octoprint.server.safe_mode,
		                             permissions=[permission.as_dict() for permission in Permissions.all()]))

		self._eventManager.fire(Events.CLIENT_OPENED, {"remoteAddress": self._remoteAddress})
		self._register()

	def on_close(self):
		self._user = self._userManager.anonymous_user_factory()
		self._groupManager.unregister_listener(self)
		self._userManager.unregister_login_status_listener(self)

		self._unregister()
		self._eventManager.fire(Events.CLIENT_CLOSED, {"remoteAddress": self._remoteAddress})

		self._logger.info("Client connection closed: %s" % self._remoteAddress)

		self._on_logout()
		self._remoteAddress = None
		self._pluginManager.unregister_message_receiver(self.on_plugin_message)

	def on_message(self, message):
		try:
			import json
			message = json.loads(message)
		except Exception:
			self._logger.warning("Invalid JSON received from client {}, ignoring: {!r}".format(self._remoteAddress, message))
			return

		if "auth" in message:
			try:
				parts = message["auth"].split(":")
				if not len(parts) == 2:
					raise ValueError()
			except ValueError:
				self._logger.warning("Got invalid auth message from client {}, ignoring: {!r}".format(self._remoteAddress, message["auth"]))
			else:
				user_id, user_session = parts
				user = self._userManager.find_user(userid=user_id, session=user_session)

				if user is not None:
					self._on_login(user)
				else:
					self._logger.warn("Unknown user/session combo: {}:{}".format(user_id, user_session))
					self._on_logout()

			self._register()

		elif "throttle" in message:
			try:
				throttle = int(message["throttle"])
				if throttle < 1:
					raise ValueError()
			except ValueError:
				self._logger.warning("Got invalid throttle factor from client {}, ignoring: {!r}".format(self._remoteAddress, message["throttle"]))
			else:
				self._throttleFactor = throttle
				self._logger.debug("Set throttle factor for client {} to {}".format(self._remoteAddress, self._throttleFactor))

	def on_printer_send_current_data(self, data):
		if not self._user.has_permission(Permissions.STATUS):
			return

		# make sure we rate limit the updates according to our throttle factor
		now = time.time()
		if now < self._lastCurrent + self._baseRateLimit * self._throttleFactor:
			return
		self._lastCurrent = now

		# add current temperature, log and message backlogs to sent data
		with self._temperatureBacklogMutex:
			temperatures = self._temperatureBacklog
			self._temperatureBacklog = []

		with self._logBacklogMutex:
			logs = self._logBacklog
			self._logBacklog = []

		with self._messageBacklogMutex:
			messages = self._messageBacklog
			self._messageBacklog = []

		busy_files = [dict(origin=v[0], path=v[1]) for v in self._fileManager.get_busy_files()]
		if "job" in data and data["job"] is not None \
				and "file" in data["job"] and "path" in data["job"]["file"] and "origin" in data["job"]["file"] \
				and data["job"]["file"]["path"] is not None and data["job"]["file"]["origin"] is not None \
				and (self._printer.is_printing() or self._printer.is_paused()):
			busy_files.append(dict(origin=data["job"]["file"]["origin"], path=data["job"]["file"]["path"]))

		data.update({
			"serverTime": time.time(),
			"temps": temperatures,
			"logs": logs,
			"messages": messages,
			"busyFiles": busy_files,
		})
		self._emit("current", payload=data)

	def on_printer_send_initial_data(self, data):
		data_to_send = dict(data)
		data_to_send["serverTime"] = time.time()
		self._emit("history", payload=data_to_send)

	def sendEvent(self, type, payload=None):
		permissions = self._event_permissions.get(type, self._event_permissions["*"])
		permissions = [x(self._user) if callable(x) else x for x in permissions]
		if not self._user or not all(map(lambda p: self._user.has_permission(p), permissions)):
			return

		processors = self._event_payload_processors.get(type, self._event_payload_processors["*"])
		for processor in processors:
			payload = processor(self._user, payload)

		self._emit("event", payload=dict(type=type, payload=payload))

	def sendTimelapseConfig(self, timelapseConfig):
		self._emit("timelapse", payload=timelapseConfig)

	def sendSlicingProgress(self, slicer, source_location, source_path, dest_location, dest_path, progress):
		self._emit("slicingProgress", payload=dict(slicer=slicer,
		                                           source_location=source_location,
		                                           source_path=source_path,
		                                           dest_location=dest_location,
		                                           dest_path=dest_path,
		                                           progress=progress))

	def sendRenderProgress(self, progress):
		self._emit("renderProgress",
		           dict(progress=progress)
		)

	def on_plugin_message(self, plugin, data, permissions=None):
		self._emit("plugin", payload=dict(plugin=plugin, data=data), permissions=permissions)

	def on_printer_add_log(self, data):
		with self._logBacklogMutex:
			self._logBacklog.append(data)

	def on_printer_add_message(self, data):
		with self._messageBacklogMutex:
			self._messageBacklog.append(data)

	def on_printer_add_temperature(self, data):
		with self._temperatureBacklogMutex:
			self._temperatureBacklog.append(data)

	def on_user_logged_out(self, user, stale=False):
		if user.get_id() == self._user.get_id() and hasattr(user, "session") and hasattr(self._user, "session") and user.session == self._user.session:
			self._logger.info("User {} logged out, logging out on socket".format(user.get_id()))
			self._on_logout()

			if stale:
				self._sendReauthRequired("stale")
			else:
				self._sendReauthRequired("logout")

	def on_user_modified(self, user):
		if user.get_id() == self._user.get_id():
			self._sendReauthRequired("modified")

	def on_user_removed(self, userid):
		if self._user.get_id() == userid:
			self._logger.info("User {} deleted, logging out on socket".format(userid))
			self._on_logout()
			self._sendReauthRequired("removed")

	def on_group_permissions_changed(self, group, added=None, removed=None):
		if self._user.is_anonymous and group == self._groupManager.guest_group:
			self._sendReauthRequired("modified")

	def on_group_subgroups_changed(self, group, added=None, removed=None):
		if self._user.is_anonymous and group == self._groupManager.guest_group:
			self._sendReauthRequired("modified")

	def _onEvent(self, event, payload):
		self.sendEvent(event, payload)

	def _register(self):
		"""Register this socket with the system if STATUS permission is available."""

		proceed = True
		for name, hook in self._register_hooks.items():
			try:
				proceed = proceed and hook(self, self._user)
			except Exception:
				self._logger.exception("Error processing register hook handler for plugin {}".format(name),
				                       extra=dict(plugin=name))

		if not proceed:
			return

		if self._registered:
			return

		if not self._user.has_permission(Permissions.STATUS):
			return

		# printer
		self._printer.register_callback(self)
		self._printer.send_initial_callback(self)

		# files
		self._fileManager.register_slicingprogress_callback(self)

		# events
		for event in octoprint.events.all_events():
			self._eventManager.subscribe(event, self._onEvent)

		# timelapse
		octoprint.timelapse.register_callback(self)
		octoprint.timelapse.notify_callback(self, timelapse=octoprint.timelapse.current)
		if octoprint.timelapse.current_render_job is not None:
			# This is a horrible hack for now to allow displaying a notification that a render job is still
			# active in the backend on a fresh connect of a client. This needs to be substituted with a proper
			# job management for timelapse rendering, analysis stuff etc that also gets cancelled when prints
			# start and so on.
			#
			# For now this is the easiest way though to at least inform the user that a timelapse is still ongoing.
			#
			# TODO remove when central job management becomes available and takes care of this for us
			self.sendEvent(Events.MOVIE_RENDERING, payload=octoprint.timelapse.current_render_job)
		self._registered = True

	def _unregister(self):
		"""Unregister this socket from the system"""

		self._printer.unregister_callback(self)
		self._fileManager.unregister_slicingprogress_callback(self)
		octoprint.timelapse.unregister_callback(self)
		for event in octoprint.events.all_events():
			self._eventManager.unsubscribe(event, self._onEvent)

	def _reregister(self):
		"""Unregister and register again"""
		self._unregister()
		self._register()

	def _sendReauthRequired(self, reason):
		self._emit("reauthRequired", payload=dict(reason=reason))

	def _emit(self, type, payload=None, permissions=None):
		proceed = True
		for name, hook in self._emit_hooks.items():
			try:
				proceed = proceed and hook(self, self._user, type, payload)
			except Exception:
				self._logger.exception("Error processing emit hook handler from plugin {}".format(name),
				                       extra=dict(plugin=name))

		if not proceed:
			return

		if permissions is None:
			permissions = self._emit_permissions.get(type, self._emit_permissions["*"])
			permissions = [x() if callable(x) else x for x in permissions]

		if not self._user or not all(map(lambda p: self._user.has_permission(p), permissions)):
			if not self._authed:
				with self._unauthed_backlog_mutex:
					if len(self._unauthed_backlog) < self._unauthed_backlog_max:
						self._unauthed_backlog.append((type, payload))
						self._logger.debug("Socket message held back until permissions cleared, added to backlog: {}".format(type))
					else:
						self._logger.debug("Socket message held back, but backlog full. Throwing message away: {}".format(type))
			return

		self._do_emit(type, payload)

	def _do_emit(self, type, payload):
		try:
			self.send({type: payload})
		except Exception as e:
			if self._logger.isEnabledFor(logging.DEBUG):
				self._logger.exception("Could not send message to client {}".format(self._remoteAddress))
			else:
				self._logger.warning("Could not send message to client {}: {}".format(self._remoteAddress, e))

	def _on_login(self, user):
		self._user = user
		self._logger.info("User {} logged in on the socket from client {}".format(user.get_name(),
		                                                                          self._remoteAddress))
		self._authed = True

		for name, hook in self._authed_hooks.items():
			try:
				hook(self, self._user)
			except Exception:
				self._logger.exception("Error processing authed hook handler for plugin {}".format(name),
				                       extra=dict(plugin=name))

		# if we have a backlog from being unauthed, process that now
		with self._unauthed_backlog_mutex:
			backlog = self._unauthed_backlog
			self._unauthed_backlog = []

		if len(backlog):
			self._logger.debug("Sending {} messages on the socket that were held back".format(len(backlog)))
			for message, payload in backlog:
				self._do_emit(message, payload)

		# trigger ClientAuthed event
		octoprint.events.eventManager().fire(octoprint.events.Events.CLIENT_AUTHED,
		                                     payload=dict(username=user.get_name(),
		                                                  remoteAddress=self._remoteAddress))

	def _on_logout(self):
		self._user = self._userManager.anonymous_user_factory()
		self._authed = False

		for name, hook in self._authed_hooks.items():
			try:
				hook(self, self._user)
			except Exception:
				self._logger.exception("Error processing authed hook handler for plugin {}".format(name),
				                       extra=dict(plugin=name))




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Marc Hannappel <salandora@gmail.com>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask import request, jsonify, abort, make_response
from werkzeug.exceptions import BadRequest
from flask_login import current_user

import octoprint.access.groups as groups
import octoprint.access.users as users

from octoprint.server import SUCCESS, groupManager, userManager
from octoprint.server.api import api, valid_boolean_trues
from octoprint.server.util.flask import no_firstrun_access
from octoprint.access.permissions import Permissions

#~~ permission api

@api.route("/access/permissions", methods=["GET"])
def get_permissions():
	return jsonify(permissions=[permission.as_dict() for permission in Permissions.all()])

#~~ group api

@api.route("/access/groups", methods=["GET"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def get_groups():
	return jsonify(groups=list(map(lambda g: g.as_dict(), groupManager.groups)))


@api.route("/access/groups", methods=["POST"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def add_group():
	if not "application/json" in request.headers["Content-Type"]:
		return make_response("Expected content-type JSON", 400)

	try:
		data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if not "key" in data:
		return make_response("Missing mandatory key field", 400)
	if not "name" in data:
		return make_response("Missing mandatory name field", 400)
	if not "permissions" in data:
		return make_response("Missing mandatory permissions field", 400)

	key = data["key"]
	name = data["name"]
	description = data.get("description", "")
	permissions = data["permissions"]
	subgroups = data["subgroups"]
	default = data.get("default", False)

	try:
		groupManager.add_group(key, name, description=description, permissions=permissions, subgroups=subgroups, default=default)
	except groups.GroupAlreadyExists:
		abort(409)
	return get_groups()


@api.route("/access/groups/<key>", methods=["GET"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def get_group(key):
	group = groupManager.find_group(key)
	if group is not None:
		return jsonify(group)
	else:
		abort(404)


@api.route("/access/groups/<key>", methods=["PUT"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def update_group(key):
	if "application/json" not in request.headers["Content-Type"]:
		return make_response("Expected content-type JSON", 400)

	try:
		data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	try:
		kwargs = dict()

		if "permissions" in data:
			kwargs["permissions"] = data["permissions"]

		if "subgroups" in data:
			kwargs["subgroups"] = data["subgroups"]

		if "default" in data:
			kwargs["default"] = data["default"] in valid_boolean_trues

		if "description" in data:
			kwargs["description"] = data["description"]

		groupManager.update_group(key, **kwargs)

		return get_groups()
	except groups.GroupCantBeChanged:
		abort(403)
	except groups.UnknownGroup:
		abort(404)


@api.route("/access/groups/<key>", methods=["DELETE"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def remove_group(key):
	try:
		groupManager.remove_group(key)
		return get_groups()
	except groups.UnknownGroup:
		abort(404)
	except groups.GroupUnremovable:
		abort(403)

#~~ user api

@api.route("/access/users", methods=["GET"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def get_users():
	if not userManager.enabled:
		return jsonify(SUCCESS)

	return jsonify(users=list(map(lambda u: u.as_dict(), userManager.get_all_users())))


@api.route("/access/users", methods=["POST"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def add_user():
	if not userManager.enabled:
		return jsonify(SUCCESS)

	if not "application/json" in request.headers["Content-Type"]:
		return make_response("Expected content-type JSON", 400)

	try:
		data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if data is None:
		return make_response("Malformed JSON body in request", 400)

	if not "name" in data:
		return make_response("Missing mandatory name field", 400)
	if not "password" in data:
		return make_response("Missing mandatory password field", 400)
	if not "active" in data:
		return make_response("Missing mandatory active field", 400)

	name = data["name"]
	password = data["password"]
	active = data["active"] in valid_boolean_trues

	groups = data.get("groups", None)
	permissions = data.get("permissions", None)

	try:
		userManager.add_user(name, password, active, permissions, groups)
	except users.UserAlreadyExists:
		abort(409)
	return get_users()


@api.route("/access/users/<username>", methods=["GET"])
@no_firstrun_access
def get_user(username):
	if not userManager.enabled:
		return jsonify(SUCCESS)

	if current_user is not None and not current_user.is_anonymous and (current_user.get_name() == username or current_user.has_permission(Permissions.ADMIN)):
		user = userManager.find_user(username)
		if user is not None:
			return jsonify(user)
		else:
			abort(404)
	else:
		abort(403)


@api.route("/access/users/<username>", methods=["PUT"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def update_user(username):
	if not userManager.enabled:
		return jsonify(SUCCESS)

	user = userManager.find_user(username)
	if user is not None:
		if not "application/json" in request.headers["Content-Type"]:
			return make_response("Expected content-type JSON", 400)

		try:
			data = request.get_json()
		except BadRequest:
			return make_response("Malformed JSON body in request", 400)


		# change groups
		if "groups" in data:
			groups = data["groups"]
			userManager.change_user_groups(username, groups)

		# change permissions
		if "permissions" in data:
			permissions = data["permissions"]
			userManager.change_user_permissions(username, permissions)

		if data is None:
			return make_response("Malformed JSON body in request", 400)

		# change activation
		if "active" in data:
			userManager.change_user_activation(username, data["active"] in valid_boolean_trues)

		return get_users()
	else:
		abort(404)


@api.route("/access/users/<username>", methods=["DELETE"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def remove_user(username):
	if not userManager.enabled:
		return jsonify(SUCCESS)

	try:
		userManager.remove_user(username)
		return get_users()
	except users.UnknownUser:
		abort(404)


@api.route("/access/users/<username>/password", methods=["PUT"])
@no_firstrun_access
def change_password_for_user(username):
	if not userManager.enabled:
		return jsonify(SUCCESS)

	if current_user is not None and not current_user.is_anonymous and (current_user.get_name() == username or current_user.is_admin):
		if not "application/json" in request.headers["Content-Type"]:
			return make_response("Expected content-type JSON", 400)

		try:
			data = request.get_json()
		except BadRequest:
			return make_response("Malformed JSON body in request", 400)

		if data is None:
			return make_response("Malformed JSON body in request", 400)

		if not "password" in data or not data["password"]:
			return make_response("password is missing from request", 400)

		try:
			userManager.change_user_password(username, data["password"])
		except users.UnknownUser:
			return make_response(("Unknown user: %s" % username, 404, []))

		return jsonify(SUCCESS)
	else:
		return make_response(("Forbidden", 403, []))


@api.route("/access/users/<username>/settings", methods=["GET"])
@no_firstrun_access
def get_settings_for_user(username):
	if not userManager.enabled:
		return jsonify(SUCCESS)

	if current_user is None or current_user.is_anonymous or (current_user.get_name() != username and not current_user.has_permission(Permissions.ADMIN)):
		return make_response("Forbidden", 403)

	try:
		return jsonify(userManager.get_all_user_settings(username))
	except users.UnknownUser:
		return make_response("Unknown user: %s" % username, 404)

@api.route("/access/users/<username>/settings", methods=["PATCH"])
@no_firstrun_access
def change_settings_for_user(username):
	if not userManager.enabled:
		return jsonify(SUCCESS)

	if current_user is None or current_user.is_anonymous or (current_user.get_name() != username and not current_user.has_permission(Permissions.ADMIN)):
		return make_response("Forbidden", 403)

	try:
		data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if data is None:
		return make_response("Malformed JSON body in request", 400)

	try:
		userManager.change_user_settings(username, data)
		return jsonify(SUCCESS)
	except users.UnknownUser:
		return make_response("Unknown user: %s" % username, 404)

@api.route("/access/users/<username>/apikey", methods=["DELETE"])
@no_firstrun_access
def delete_apikey_for_user(username):
	if not userManager.enabled:
		return jsonify(SUCCESS)

	if current_user is not None and not current_user.is_anonymous and (current_user.get_name() == username or current_user.has_permission(Permissions.ADMIN)):
		try:
			userManager.delete_api_key(username)
		except users.UnknownUser:
			return make_response(("Unknown user: %s" % username, 404, []))
		return jsonify(SUCCESS)
	else:
		return make_response(("Forbidden", 403, []))


@api.route("/access/users/<username>/apikey", methods=["POST"])
@no_firstrun_access
def generate_apikey_for_user(username):
	if not userManager.enabled:
		return jsonify(SUCCESS)

	if current_user is not None and not current_user.is_anonymous and (current_user.get_name() == username or current_user.has_permission(Permissions.ADMIN)):
		try:
			apikey = userManager.generate_api_key(username)
		except users.UnknownUser:
			return make_response(("Unknown user: %s" % username, 404, []))
		return jsonify({"apikey": apikey})
	else:
		return make_response(("Forbidden", 403, []))

def _to_external_permissions(*permissions):
	return list(map(lambda p: p.get_name(), permissions))

def _to_external_groups(*groups):
	return list(map(lambda g: g.get_name(), groups))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

"""
The SSDP/UPNP implementations has been largely inspired by https://gist.github.com/schlamar/2428250
"""

import logging
import flask
from flask_babel import gettext

# noinspection PyCompatibility
from builtins import range

import octoprint.plugin
import octoprint.util

try:
	# TODO: looks like pybonjour is py2 only - find alternatives or port?
	import pybonjour
except ImportError:
	pybonjour = False


def __plugin_load__():
	if not pybonjour:
		# no pybonjour available, we can't use that
		logging.getLogger("octoprint.plugins." + __name__).info("pybonjour is not installed, Zeroconf Discovery won't be available")

	plugin = DiscoveryPlugin()

	global __plugin_implementation__
	__plugin_implementation__ = plugin

	global __plugin_helpers__
	__plugin_helpers__ = dict(
		ssdp_browse=plugin.ssdp_browse
	)
	if pybonjour:
		__plugin_helpers__.update(dict(
			zeroconf_browse=plugin.zeroconf_browse,
			zeroconf_register=plugin.zeroconf_register,
			zeroconf_unregister=plugin.zeroconf_unregister
		))

class DiscoveryPlugin(octoprint.plugin.StartupPlugin,
                      octoprint.plugin.ShutdownPlugin,
                      octoprint.plugin.BlueprintPlugin,
                      octoprint.plugin.SettingsPlugin):

	ssdp_multicast_addr = "239.255.255.250"

	ssdp_multicast_port = 1900

	# noinspection PyMissingConstructor
	def __init__(self):
		self.host = None
		self.port = None

		# zeroconf
		self._sd_refs = dict()
		self._cnames = dict()

		# upnp/ssdp
		self._ssdp_monitor_active = False
		self._ssdp_monitor_thread = None
		self._ssdp_notify_timeout = 10
		self._ssdp_last_notify = 0

	##~~ SettingsPlugin API

	def get_settings_defaults(self):
		return {
			"publicHost": None,
			"publicPort": None,
			"pathPrefix": None,
			"httpUsername": None,
			"httpPassword": None,
			"upnpUuid": None,
			"zeroConf": [],
			"model": {
				"name": None,
				"description": None,
				"number": None,
				"url": None,
				"serial": None,
				"vendor": None,
				"vendorUrl": None
			}
		}

	##~~ BlueprintPlugin API -- used for providing the SSDP device descriptor XML

	@octoprint.plugin.BlueprintPlugin.route("/discovery.xml", methods=["GET"])
	def discovery(self):
		self._logger.debug("Rendering discovery.xml")

		modelName = self._settings.get(["model", "name"])
		if not modelName:
			import octoprint.server
			modelName = octoprint.server.DISPLAY_VERSION

		vendor = self._settings.get(["model", "vendor"])
		vendorUrl = self._settings.get(["model", "vendorUrl"])
		if not vendor:
			vendor = "The OctoPrint Project"
			vendorUrl = "http://www.octoprint.org/"

		response = flask.make_response(flask.render_template("discovery.xml.jinja2",
		                                                     friendlyName=self.get_instance_name(),
		                                                     manufacturer=vendor,
		                                                     manufacturerUrl=vendorUrl,
		                                                     modelName=modelName,
		                                                     modelDescription=self._settings.get(["model", "description"]),
		                                                     modelNumber=self._settings.get(["model", "number"]),
		                                                     modelUrl=self._settings.get(["model", "url"]),
		                                                     serialNumber=self._settings.get(["model", "serial"]),
		                                                     uuid=self.get_uuid(),
		                                                     presentationUrl=flask.url_for("index", _external=True)))
		response.headers['Content-Type'] = 'application/xml'
		return response

	def is_blueprint_protected(self):
		return False

	##~~ StartupPlugin API -- used for registering OctoPrint's Zeroconf and SSDP services upon application startup

	def on_startup(self, host, port):
		public_host = self._settings.get(["publicHost"])
		if public_host:
			host = public_host
		public_port = self._settings.get(["publicPort"])
		if public_port:
			port = public_port

		self.host = host
		self.port = port

		# Zeroconf
		self.zeroconf_register("_http._tcp", self.get_instance_name(), txt_record=self._create_http_txt_record_dict())
		self.zeroconf_register("_octoprint._tcp", self.get_instance_name(), txt_record=self._create_octoprint_txt_record_dict())
		for zeroconf in self._settings.get(["zeroConf"]):
			if "service" in zeroconf:
				self.zeroconf_register(
					zeroconf["service"],
					zeroconf["name"] if "name" in zeroconf else self.get_instance_name(),
					port=zeroconf["port"] if "port" in zeroconf else None,
					txt_record=zeroconf["txtRecord"] if "txtRecord" in zeroconf else None
				)

		# SSDP
		self._ssdp_register()

	##~~ ShutdownPlugin API -- used for unregistering OctoPrint's Zeroconf and SSDP service upon application shutdown

	def on_shutdown(self):
		for key in self._sd_refs:
			reg_type, port = key
			self.zeroconf_unregister(reg_type, port)

		self._ssdp_unregister()

	##~~ helpers

	# ZeroConf

	def zeroconf_register(self, reg_type, name=None, port=None, txt_record=None):
		"""
		Registers a new service with Zeroconf/Bonjour/Avahi.

		:param reg_type: type of service to register, e.g. "_gntp._tcp"
		:param name: displayable name of the service, if not given defaults to the OctoPrint instance name
		:param port: port to register for the service, if not given defaults to OctoPrint's (public) port
		:param txt_record: optional txt record to attach to the service, dictionary of key-value-pairs
		"""

		if not pybonjour:
			return

		if not name:
			name = self.get_instance_name()
		if not port:
			port = self.port

		params = dict(
			name=name,
			regtype=reg_type,
			port=port
		)
		if txt_record:
			params["txtRecord"] = pybonjour.TXTRecord(txt_record)

		key = (reg_type, port)

		counter = 1
		while True:
			try:
				self._sd_refs[key] = pybonjour.DNSServiceRegister(**params)
				self._logger.info("Registered '{name}' for {regtype}".format(**params))
				return True
			except pybonjour.BonjourError as be:
				if be.errorCode == pybonjour.kDNSServiceErr_NameConflict:
					# Name already registered by different service, let's try a counter postfix. See #2852
					counter += 1
					params["name"] = "{} ({})".format(name, counter)
				else:
					raise

	def zeroconf_unregister(self, reg_type, port=None):
		"""
		Unregisters a previously registered Zeroconf/Bonjour/Avahi service identified by service and port.

		:param reg_type: the type of the service to be unregistered
		:param port: the port of the service to be unregistered, defaults to OctoPrint's (public) port if not given
		:return:
		"""

		if not pybonjour:
			return

		if not port:
			port = self.port

		key = (reg_type, port)
		if not key in self._sd_refs:
			return

		sd_ref = self._sd_refs[key]
		try:
			sd_ref.close()
			self._logger.debug("Unregistered {reg_type} on port {port}".format(reg_type=reg_type, port=port))
		except Exception:
			self._logger.exception("Could not unregister {reg_type} on port {port}".format(reg_type=reg_type, port=port))

	def zeroconf_browse(self, service_type, block=True, callback=None, browse_timeout=5, resolve_timeout=5):
		"""
		Browses for services on the local network providing the specified service type. Can be used either blocking or
		non-blocking.

		The non-blocking version (default behaviour) will not return until the lookup has completed and
		return all results that were found.

		For non-blocking version, set `block` to `False` and provide a `callback` to be called once the lookup completes.
		If no callback is provided in non-blocking mode, a ValueError will be raised.

		The results are provided as a list of discovered services, with each service being described by a dictionary
		with the following keys:

		  * `name`: display name of the service
		  * `host`: host name of the service
		  * `post`: port the service is listening on
		  * `txt_record`: TXT record of the service as a dictionary, exact contents depend on the service

		Callbacks will be called with that list as the single parameter supplied to them. Thus, the following is an
		example for a valid callback:

		    def browse_callback(results):
		      for result in results:
		        print "Name: {name}, Host: {host}, Port: {port}, TXT: {txt_record!r}".format(**result)

		:param service_type: the service type to browse for
		:param block: whether to block, defaults to True
		:param callback: callback to call once lookup has completed, must be set when `block` is set to `False`
		:param browse_timeout: timeout for browsing operation
		:param resolve_timeout: timeout for resolving operations for discovered records
		:return: if `block` is `True` a list of the discovered services, an empty list otherwise (results will then be
		         supplied to the callback instead)
		"""

		if not pybonjour:
			return None

		import threading
		import select

		if not block and not callback:
			raise ValueError("Non-blocking mode but no callback given")

		result = []
		result_available = threading.Event()
		result_available.clear()

		resolved = []

		def resolve_callback(sd_ref, flags, interface_index, error_code, fullname, hosttarget, port, txt_record):
			if error_code == pybonjour.kDNSServiceErr_NoError:
				txt_record_dict = None
				if txt_record:
					record = pybonjour.TXTRecord.parse(txt_record)
					txt_record_dict = dict()
					for key, value in record:
						txt_record_dict[key] = value

				name = fullname[:fullname.find(service_type) - 1].replace("\\032", " ")
				host = hosttarget[:-1]

				self._logger.debug("Resolved a result for Zeroconf resolution of {service_type}: {name} @ {host}".format(service_type=service_type, name=name, host=host))
				result.append(dict(
					name=name,
					host=host,
					port=port,
					txt_record=txt_record_dict
				))
				resolved.append(True)

		def browse_callback(sd_ref, flags, interface_index, error_code, service_name, regtype, reply_domain):
			if error_code != pybonjour.kDNSServiceErr_NoError:
				return

			if not (flags & pybonjour.kDNSServiceFlagsAdd):
				return

			self._logger.debug("Got a browsing result for Zeroconf resolution of {service_type}, resolving...".format(service_type=service_type))
			resolve_ref = pybonjour.DNSServiceResolve(0, interface_index, service_name, regtype, reply_domain, resolve_callback)

			try:
				while not resolved:
					ready = select.select([resolve_ref], [], [], resolve_timeout)
					if resolve_ref not in ready[0]:
						break

					pybonjour.DNSServiceProcessResult(resolve_ref)
				else:
					resolved.pop()
			finally:
				resolve_ref.close()

		self._logger.debug("Browsing Zeroconf for {service_type}".format(service_type=service_type))

		def browse():
			sd_ref = pybonjour.DNSServiceBrowse(regtype=service_type, callBack=browse_callback)
			try:
				while True:
					ready = select.select([sd_ref], [], [], browse_timeout)

					if not ready[0]:
						break

					if sd_ref in ready[0]:
						pybonjour.DNSServiceProcessResult(sd_ref)
			finally:
				sd_ref.close()

			if callback:
				callback(result)
			result_available.set()

		browse_thread = threading.Thread(target=browse)
		browse_thread.daemon = True
		browse_thread.start()

		if block:
			result_available.wait()
			return result
		else:
			return []

	# SSDP/UPNP

	def ssdp_browse(self, query, block=True, callback=None, timeout=1, retries=5):
		"""
		Browses for UPNP services matching the supplied query. Can be used either blocking or
		non-blocking.

		The non-blocking version (default behaviour) will not return until the lookup has completed and
		return all results that were found.

		For non-blocking version, set `block` to `False` and provide a `callback` to be called once the lookup completes.
		If no callback is provided in non-blocking mode, a ValueError will be raised.

		The results are provided as a list of discovered locations of device descriptor files.

		Callbacks will be called with that list as the single parameter supplied to them. Thus, the following is an
		example for a valid callback:

		    def browse_callback(results):
		      for result in results:
		        print("Location: {}".format(result))

		:param query: the SSDP query to send, e.g. "upnp:rootdevice" to search for all devices
		:param block: whether to block, defaults to True
		:param callback: callback to call in non-blocking mode when lookup has finished, must be set if block is False
		:param timeout: timeout in seconds to wait for replies to the M-SEARCH query per interface, defaults to 1
		:param retries: number of retries to perform the lookup on all interfaces, defaults to 5
		:return: if `block` is `True` a list of the discovered devices, an empty list otherwise (results will then be
		         supplied to the callback instead)
		"""

		import threading
		import io

		try:
			# noinspection PyCompatibility
			from http.client import HTTPResponse # py3
		except ImportError:
			# noinspection PyCompatibility
			from httplib import HTTPResponse # py2

		class Response(HTTPResponse):
			def __init__(self, response_text):
				self.fp = io.BytesIO(response_text)
				self.debuglevel = 0
				self.strict = 0
				self.msg = None
				self._method = None
				self.begin()

		result = []
		result_available = threading.Event()
		result_available.clear()

		def browse():
			import socket

			socket.setdefaulttimeout(timeout)

			search_message = "".join([
				"M-SEARCH * HTTP/1.1\r\n",
				"ST: {query}\r\n",
				"MX: 3\r\n",
				"MAN: \"ssdp:discovery\"\r\n",
				"HOST: {mcast_addr}:{mcast_port}\r\n\r\n"
			])

			for _ in range(retries):
				for addr in octoprint.util.interface_addresses():
					try:
						sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
						sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
						sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
						sock.bind((addr, 0))

						message = search_message.format(query=query,
						                                mcast_addr=self.__class__.ssdp_multicast_addr,
						                                mcast_port=self.__class__.ssdp_multicast_port)
						for _ in range(2):
							sock.sendto(message.encode("utf-8"), (self.__class__.ssdp_multicast_addr, self.__class__.ssdp_multicast_port))

						try:
							data = sock.recv(1024)
						except socket.timeout:
							pass
						else:
							response = Response(data)

							result.append(response.getheader("Location"))
					except Exception:
						pass

			if callback:
				callback(result)
			result_available.set()

		browse_thread = threading.Thread(target=browse)
		browse_thread.daemon = True
		browse_thread.start()

		if block:
			result_available.wait()
			return result
		else:
			return []

	##~~ internals

	# Zeroconf

	def _create_http_txt_record_dict(self):
		"""
		Creates a TXT record for the _http._tcp Zeroconf service supplied by this OctoPrint instance.

		Defines the keys for _http._tcp as defined in http://www.dns-sd.org/txtrecords.html

		:return: a dictionary containing the defined key-value-pairs, ready to be turned into a TXT record
		"""

		# determine path entry
		path = "/"
		if self._settings.get(["pathPrefix"]):
			path = self._settings.get(["pathPrefix"])
		else:
			prefix = self._settings.global_get(["server", "reverseProxy", "prefixFallback"])
			if prefix:
				path = prefix

		# fetch username and password (if set)
		username = self._settings.get(["httpUsername"])
		password = self._settings.get(["httpPassword"])

		entries = dict(
			path=path
		)

		if username and password:
			entries.update(dict(u=username, p=password))

		return entries

	def _create_octoprint_txt_record_dict(self):
		"""
		Creates a TXT record for the _octoprint._tcp Zeroconf service supplied by this OctoPrint instance.

		The following keys are defined:

		  * `path`: path prefix to actual OctoPrint instance, inherited from _http._tcp
		  * `u`: username if HTTP Basic Auth is used, optional, inherited from _http._tcp
		  * `p`: password if HTTP Basic Auth is used, optional, inherited from _http._tcp
		  * `version`: OctoPrint software version
		  * `api`: OctoPrint API version
		  * `model`: Model of the device that is running OctoPrint
		  * `vendor`: Vendor of the device that is running OctoPrint

		:return: a dictionary containing the defined key-value-pairs, ready to be turned into a TXT record
		"""

		entries = self._create_http_txt_record_dict()

		import octoprint.server
		import octoprint.server.api

		entries.update(dict(
			version=octoprint.server.VERSION,
			api=octoprint.server.api.VERSION,
			))

		modelName = self._settings.get(["model", "name"])
		if modelName:
			entries.update(dict(model=modelName))
		vendor = self._settings.get(["model", "vendor"])
		if vendor:
			entries.update(dict(vendor=vendor))

		return entries

	# SSDP/UPNP

	def _ssdp_register(self):
		"""
		Registers the OctoPrint instance as basic service with a presentation URL pointing to the web interface
		"""

		import threading

		self._ssdp_monitor_active = True

		self._ssdp_monitor_thread = threading.Thread(target=self._ssdp_monitor, kwargs=dict(timeout=self._ssdp_notify_timeout))
		self._ssdp_monitor_thread.daemon = True
		self._ssdp_monitor_thread.start()

	def _ssdp_unregister(self):
		"""
		Unregisters the OctoPrint instance again
		"""

		self._ssdp_monitor_active = False
		if self.host and self.port:
			for _ in range(2):
				self._ssdp_notify(alive=False)

	def _ssdp_notify(self, alive=True):
		"""
		Sends an SSDP notify message across the connected networks.

		:param alive: True to send an "ssdp:alive" message, False to send an "ssdp:byebye" message
		"""

		import socket

		if alive and self._ssdp_last_notify + self._ssdp_notify_timeout > octoprint.util.monotonic_time():
			# we just sent an alive, no need to send another one now
			return

		if alive and not self._ssdp_monitor_active:
			# the monitor already shut down, alive messages don't make sense anymore as byebye will shortly follow
			return

		for addr in octoprint.util.interface_addresses():
			try:
				sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
				sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
				sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
				sock.bind((addr, 0))

				location = "http://{addr}:{port}/plugin/discovery/discovery.xml".format(addr=addr, port=self.port)

				self._logger.debug("Sending NOTIFY {} via {}".format("alive" if alive else "byebye", addr))
				notify_message = "".join([
					"NOTIFY * HTTP/1.1\r\n",
					"Server: Python/2.7\r\n",
					"Cache-Control: max-age=900\r\n",
					"Location: {location}\r\n",
					"NTS: {nts}\r\n",
					"NT: upnp:rootdevice\r\n",
					"USN: uuid:{uuid}::upnp:rootdevice\r\n",
					"HOST: {mcast_addr}:{mcast_port}\r\n\r\n"
				])
				message = notify_message.format(uuid=self.get_uuid(),
				                                location=location,
				                                nts="ssdp:alive" if alive else "ssdp:byebye",
				                                mcast_addr=self.__class__.ssdp_multicast_addr,
				                                mcast_port=self.__class__.ssdp_multicast_port)
				for _ in range(2):
					# send twice, stuff might get lost, it's only UDP
					sock.sendto(message.encode("utf-8"), (self.__class__.ssdp_multicast_addr, self.__class__.ssdp_multicast_port))
			except Exception:
				pass

		self._ssdp_last_notify = octoprint.util.monotonic_time()

	def _ssdp_monitor(self, timeout=5):
		"""
		Monitor thread that listens on the multicast address for M-SEARCH requests and answers them if they are relevant

		:param timeout: timeout after which to stop waiting for M-SEARCHs for a short while in order to put out an
		                alive message
		"""

		try:
			# noinspection PyCompatibility
			from http.server import BaseHTTPRequestHandler
		except ImportError:
			# noinspection PyCompatibility
			from BaseHTTPServer import BaseHTTPRequestHandler

		from io import BytesIO
		import socket

		socket.setdefaulttimeout(timeout)

		location_message = "".join([
			"HTTP/1.1 200 OK\r\n",
			"ST: upnp:rootdevice\r\n",
			"USN: uuid:{uuid}::upnp:rootdevice\r\n",
			"Location: {location}\r\n",
			"Cache-Control: max-age=60\r\n\r\n"
		])

		class Request(BaseHTTPRequestHandler):

			# noinspection PyMissingConstructor
			def __init__(self, request_text):
				self.rfile = BytesIO(request_text)
				self.raw_requestline = self.rfile.readline()
				self.error_code = self.error_message = None
				self.parse_request()

			def send_error(self, code, message=None):
				self.error_code = code
				self.error_message = message

		sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
		sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
		sock.bind(('', self.__class__.ssdp_multicast_port))

		sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, socket.inet_aton(self.__class__.ssdp_multicast_addr) + socket.inet_aton('0.0.0.0'))

		self._logger.info("Registered {} for SSDP".format(self.get_instance_name()))

		self._ssdp_notify(alive=True)

		try:
			while self._ssdp_monitor_active:
				try:
					data, address = sock.recvfrom(4096)
					request = Request(data)
					if not request.error_code and request.command == "M-SEARCH" and request.path == "*" and (request.headers["ST"] == "upnp:rootdevice" or request.headers["ST"] == "ssdp:all") and request.headers["MAN"] == '"ssdp:discover"':
						interface_address = octoprint.util.address_for_client(*address)
						if not interface_address:
							self._logger.warning("Can't determine address to user for client {}, not sending a M-SEARCH reply".format(address))
							continue
						message = location_message.format(uuid=self.get_uuid(),
						                                  location="http://{host}:{port}/plugin/discovery/discovery.xml".format(host=interface_address, port=self.port))
						sock.sendto(message.encode("utf-8"), address)
						self._logger.debug("Sent M-SEARCH reply for {path} and {st} to {address!r}".format(path=request.path, st=request.headers["ST"], address=address))
				except socket.timeout:
					pass
				finally:
					self._ssdp_notify(alive=True)
		finally:
			try:
				sock.close()
			except Exception:
				pass

	##~~ helpers

	def get_uuid(self):
		upnpUuid = self._settings.get(["upnpUuid"])
		if upnpUuid is None:
			import uuid
			upnpUuid = str(uuid.uuid4())
			self._settings.set(["upnpUuid"], upnpUuid)
			self._settings.save()
		return upnpUuid

	def get_instance_name(self):
		name = self._settings.global_get(["appearance", "name"])
		if name:
			return "OctoPrint instance \"{}\"".format(name)
		else:
			import socket
			return "OctoPrint instance on {}".format(socket.gethostname())


__plugin_name__ = "Discovery"
__plugin_author__ = "Gina Häußge"
__plugin_url__ = "http://docs.octoprint.org/en/master/bundledplugins/discovery.html"
__plugin_description__ = "Makes the OctoPrint instance discoverable via Bonjour/Avahi/Zeroconf and uPnP"
__plugin_disabling_discouraged__ = gettext("Without this plugin your OctoPrint instance will no longer be "
                                           "discoverable on the network via Bonjour and uPnP.")
__plugin_license__ = "AGPLv3"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin

from flask_babel import gettext

import requests
import hashlib
import logging

try:
	# noinspection PyCompatibility
	from urllib.parse import urlencode
except ImportError:
	from urllib import urlencode

# noinspection PyCompatibility
import concurrent.futures

from octoprint.util import RepeatedTimer, monotonic_time
from octoprint.util.version import get_octoprint_version_string
from octoprint.events import Events

TRACKING_URL = "https://tracking.octoprint.org/track/{id}/{event}/"

# noinspection PyMissingConstructor
class TrackingPlugin(octoprint.plugin.SettingsPlugin,
                     octoprint.plugin.EnvironmentDetectionPlugin,
                     octoprint.plugin.StartupPlugin,
                     octoprint.plugin.ShutdownPlugin,
                     octoprint.plugin.TemplatePlugin,
                     octoprint.plugin.AssetPlugin,
                     octoprint.plugin.WizardPlugin,
                     octoprint.plugin.EventHandlerPlugin):

	def __init__(self):
		self._environment = None
		self._throttle_state = None
		self._helpers_get_throttle_state = None
		self._printer_connection_parameters = None
		self._url = None
		self._ping_worker = None
		self._pong_worker = None
		self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)

		self._record_next_firmware_info = False

		self._startup_time = monotonic_time()

	def initialize(self):
		self._init_id()

	##~~ SettingsPlugin

	def get_settings_defaults(self):
		return dict(enabled=None,
		            unique_id=None,
		            server=TRACKING_URL,
		            ping=15*60,
		            pong=24*60*60,
		            events=dict(pong=True,
		                        startup=True,
		                        printjob=True,
		                        commerror=True,
		                        plugin=True,
		                        update=True,
		                        printer=True,
		                        printer_safety_check=True,
		                        throttled=True,
		                        slicing=True))

	def get_settings_restricted_paths(self):
		return dict(admin=[["enabled"], ["unique_id"], ["events"]],
		            never=[["server"], ["ping"]])

	def on_settings_save(self, data):
		enabled = self._settings.get(["enabled"])

		octoprint.plugin.SettingsPlugin.on_settings_save(self, data)

		if enabled is None and self._settings.get(["enabled"]):
			# tracking was just enabled, let's start up tracking
			self._start_tracking()

	##~~ EnvironmentDetectionPlugin

	def on_environment_detected(self, environment, *args, **kwargs):
		self._environment = environment

	##~~ StartupPlugin

	def on_after_startup(self):
		self._start_tracking()

	##~~ ShutdownPlugin

	def on_shutdown(self):
		if not self._settings.get_boolean(["enabled"]):
			return
		self._track_shutdown()

	##~~ EventHandlerPlugin

	# noinspection PyUnresolvedReferences
	def on_event(self, event, payload):
		if not self._settings.get_boolean(["enabled"]):
			return

		if event in (Events.PRINT_STARTED, Events.PRINT_DONE, Events.PRINT_FAILED, Events.PRINT_CANCELLED):
			self._track_printjob_event(event, payload)

		elif event in (Events.ERROR,):
			self._track_commerror_event(event, payload)

		elif event in (Events.CONNECTED,):
			self._printer_connection_parameters = dict(port=payload["port"],
			                                           baudrate=payload["baudrate"])
			self._record_next_firmware_info = True

		elif event in (Events.FIRMWARE_DATA,) and self._record_next_firmware_info:
			self._record_next_firmware_info = False
			self._track_printer_event(event, payload)

		elif event in (Events.SLICING_STARTED,):
			self._track_slicing_event(event, payload)

		elif hasattr(Events, "PLUGIN_PLUGINMANAGER_INSTALL_PLUGIN") and \
			event in (Events.PLUGIN_PLUGINMANAGER_INSTALL_PLUGIN, Events.PLUGIN_PLUGINMANAGER_UNINSTALL_PLUGIN,
			          Events.PLUGIN_PLUGINMANAGER_ENABLE_PLUGIN, Events.PLUGIN_PLUGINMANAGER_DISABLE_PLUGIN):
			self._track_plugin_event(event, payload)

		elif hasattr(Events, "PLUGIN_SOFTWAREUPDATE_UPDATE_SUCCEEDED") and \
			event in (Events.PLUGIN_SOFTWAREUPDATE_UPDATE_SUCCEEDED, Events.PLUGIN_SOFTWAREUPDATE_UPDATE_FAILED):
			self._track_update_event(event, payload)

		elif hasattr(Events, "PLUGIN_PI_SUPPORT_THROTTLE_STATE") and event in (Events.PLUGIN_PI_SUPPORT_THROTTLE_STATE,):
			self._throttle_state = payload
			self._track_throttle_event(event, payload)

		elif hasattr(Events, "PLUGIN_PRINTER_SAFETY_CHECK_WARNING") and event in (Events.PLUGIN_PRINTER_SAFETY_CHECK_WARNING,):
			self._track_printer_safety_event(event, payload)

	##~~ TemplatePlugin

	def get_template_configs(self):
		return [
			dict(type="settings", name=gettext("Anonymous Usage Tracking"), template="tracking_settings.jinja2", custom_bindings=False),
			dict(type="wizard", name=gettext("Anonymous Usage Tracking"), template="tracking_wizard.jinja2", custom_bindings=True, mandatory=True)
		]

	##~~ AssetPlugin

	def get_assets(self):
		return dict(js=["js/usage.js"])

	##~~ WizardPlugin

	def is_wizard_required(self):
		return self._settings.get(["enabled"]) is None

	##~~ helpers

	def _init_id(self):
		if not self._settings.get(["unique_id"]):
			import uuid
			self._settings.set(["unique_id"], str(uuid.uuid4()))
			self._settings.save()

	def _start_tracking(self):
		if not self._settings.get_boolean(["enabled"]):
			return

		if self._ping_worker is None:
			ping_interval = self._settings.get_int(["ping"])
			if ping_interval:
				self._ping_worker = RepeatedTimer(ping_interval, self._track_ping, run_first=True)
				self._ping_worker.start()

		if self._pong_worker is None:
			pong_interval = self._settings.get(["pong"])
			if pong_interval:
				self._pong_worker = RepeatedTimer(pong_interval, self._track_pong, run_first=True)
				self._pong_worker.start()

		if self._helpers_get_throttle_state is None:
			# cautiously look for the get_throttled helper from pi_support
			pi_helper = self._plugin_manager.get_helpers("pi_support", "get_throttled")
			if pi_helper and 'get_throttled' in pi_helper:
				self._helpers_get_throttle_state = pi_helper['get_throttled']

		# now that we have everything set up, phone home.
		self._track_startup()

	def _track_ping(self):
		if not self._settings.get_boolean(["enabled"]):
			return

		uptime = int(monotonic_time() - self._startup_time)
		self._track("ping", octoprint_uptime=uptime)

	def _track_pong(self):
		if not self._settings.get_boolean(["events", "pong"]):
			return

		plugins = self._plugin_manager.enabled_plugins
		plugins_thirdparty = [plugin for plugin in plugins.values() if not plugin.bundled]
		payload = dict(plugins=",".join(map(lambda x: "{}:{}".format(x.key.lower(),
		                                                             x.version.lower() if x.version else "?"),
		                                    plugins_thirdparty)))

		self._track("pong", body=True, **payload)

	def _track_startup(self):
		if not self._settings.get_boolean(["events", "startup"]):
			return

		payload = dict(version=get_octoprint_version_string(),
		               os=self._environment["os"]["id"],
		               python=self._environment["python"]["version"],
		               pip=self._environment["python"]["pip"],
		               cores=self._environment["hardware"]["cores"],
		               freq=self._environment["hardware"]["freq"],
		               ram=self._environment["hardware"]["ram"])

		if "plugins" in self._environment and "pi_support" in self._environment["plugins"]:
			payload["pi_model"] = self._environment["plugins"]["pi_support"]["model"]

			if "octopi_version" in self._environment["plugins"]["pi_support"]:
				payload["octopi_version"] = self._environment["plugins"]["pi_support"]["octopi_version"]

		self._track("startup", **payload)

	def _track_shutdown(self):
		if not self._settings.get_boolean(["enabled"]):
			return

		if not self._settings.get_boolean(["events", "startup"]):
			return

		self._track("shutdown")

	def _track_plugin_event(self, event, payload):
		if not self._settings.get_boolean(["events", "plugin"]):
			return

		if event.endswith("_install_plugin"):
			self._track("install_plugin", plugin=payload.get("id"), plugin_version=payload.get("version"))
		elif event.endswith("_uninstall_plugin"):
			self._track("uninstall_plugin", plugin=payload.get("id"), plugin_version=payload.get("version"))
		elif event.endswith("_enable_plugin"):
			self._track("enable_plugin", plugin=payload.get("id"), plugin_version=payload.get("version"))
		elif event.endswith("_disable_plugin"):
			self._track("disable_plugin", plugin=payload.get("id"), plugin_version=payload.get("version"))

	def _track_update_event(self, event, payload):
		if not self._settings.get_boolean(["events", "update"]):
			return

		if event.endswith("_update_succeeded"):
			self._track("update_successful", target=payload.get("target"), from_version=payload.get("from_version"), to_version=payload.get("to_version"))
		elif event.endswith("_update_failed"):
			self._track("update_failed", target=payload.get("target"), from_version=payload.get("from_version"), to_version=payload.get("to_version"))

	def _track_throttle_event(self, event, payload):
		if not self._settings.get_boolean(["events", "throttled"]):
			return

		args = dict(throttled_now=payload["current_issue"],
		            throttled_past=payload["past_issue"],
		            throttled_mask=payload["raw_value"],
		            throttled_voltage_now=payload["current_undervoltage"],
		            throttled_voltage_past=payload["past_undervoltage"],
		            throttled_overheat_now=payload["current_overheat"],
		            throttled_overheat_past=payload["past_overheat"])

		if payload["current_issue"]:
			track_event = "system_throttled"
		else:
			track_event = "system_unthrottled"

		if track_event is not None:
			self._track(track_event, **args)

	def _track_commerror_event(self, event, payload):
		if not self._settings.get_boolean(["events", "commerror"]):
			return

		if not "reason" in payload or not "error" in payload:
			return

		track_event = "commerror_{}".format(payload["reason"])
		args = dict(commerror_text=payload["error"])

		if callable(self._helpers_get_throttle_state):
			try:
				throttle_state = self._helpers_get_throttle_state(run_now=True)
				if throttle_state and (throttle_state.get("current_issue", False) or throttle_state.get("past_issue", False)):
					args["throttled_now"] = throttle_state["current_issue"]
					args["throttled_past"] = throttle_state["past_issue"]
					args["throttled_mask"] = throttle_state["raw_value"]
			except Exception:
				# ignored
				pass

		self._track(track_event, **args)

	def _track_printjob_event(self, event, payload):
		if not self._settings.get_boolean(["events", "printjob"]):
			return

		unique_id = self._settings.get(["unique_id"])
		if not unique_id:
			return

		sha = hashlib.sha1()
		sha.update(payload.get("path").encode("utf-8"))
		sha.update(unique_id.encode("utf-8"))

		track_event = None
		args = dict(origin=payload.get("origin"), file=sha.hexdigest())

		if event == Events.PRINT_STARTED:
			track_event = "print_started"
		elif event == Events.PRINT_DONE:
			try:
				elapsed = int(payload.get("time", 0))
				if elapsed:
					args["elapsed"] = elapsed
			except (ValueError, TypeError):
				pass
			track_event = "print_done"
		elif event == Events.PRINT_FAILED:
			try:
				elapsed = int(payload.get("time", 0))
				if elapsed:
					args["elapsed"] = elapsed
			except (ValueError, TypeError):
				pass
			args["reason"] = payload.get("reason", "unknown")

			if "error" in payload and self._settings.get_boolean(["events", "commerror"]):
				args["commerror_text"] = payload["error"]

			track_event = "print_failed"
		elif event == Events.PRINT_CANCELLED:
			try:
				elapsed = int(payload.get("time", 0))
				if elapsed:
					args["elapsed"] = elapsed
			except (ValueError, TypeError):
				pass
			track_event = "print_cancelled"

		if callable(self._helpers_get_throttle_state):
			try:
				throttle_state = self._helpers_get_throttle_state(run_now=True)
				if throttle_state and (throttle_state.get("current_issue", False) or throttle_state.get("past_issue", False)):
					args["throttled_now"] = throttle_state["current_issue"]
					args["throttled_past"] = throttle_state["past_issue"]
					args["throttled_mask"] = throttle_state["raw_value"]
			except Exception:
				# ignored
				pass

		if track_event is not None:
			self._track(track_event, **args)

	def _track_printer_event(self, event, payload):
		if not self._settings.get_boolean(["events", "printer"]):
			return

		if event in (Events.FIRMWARE_DATA,):
			args = dict(firmware_name=payload["name"])
			if self._printer_connection_parameters:
				args["printer_port"] = self._printer_connection_parameters["port"]
				args["printer_baudrate"] = self._printer_connection_parameters["baudrate"]
			self._track("printer_connected", **args)

	def _track_printer_safety_event(self, event, payload):
		if not self._settings.get_boolean(["events", "printer_safety_check"]):
			return

		self._track("printer_safety_warning",
		            printer_safety_warning_type=payload.get("warning_type", "unknown"),
		            printer_safety_check_name=payload.get("check_name", "unknown"))

	def _track_slicing_event(self, event, payload):
		if not self._settings.get_boolean(["events", "slicing"]):
			return

		self._track("slicing_started",
		            slicer=payload.get(b"slicer", "unknown"))

	def _track(self, event, **kwargs):
		if not self._settings.get_boolean(["enabled"]):
			return

		self._executor.submit(self._do_track, event, **kwargs)

	def _do_track(self, event, body=False, **kwargs):
		if not self._connectivity_checker.online:
			return

		if not self._settings.get_boolean(["enabled"]):
			return

		unique_id = self._settings.get(["unique_id"])
		if not unique_id:
			return

		server = self._settings.get(["server"])
		url = server.format(id=unique_id, event=event)
		# Don't print the URL or UUID! That would expose the UUID in forums/tickets
		# if pasted. It's okay for the user to know their uuid, but it shouldn't be shared.

		headers = {"User-Agent": "OctoPrint/{}".format(get_octoprint_version_string())}
		try:
			params = urlencode(kwargs, doseq=True).replace("+", "%20")

			if body:
				requests.post(url,
				              data=params,
				              timeout=3.1,
				              headers=headers)
			else:
				requests.get(url,
				             params=params,
				             timeout=3.1,
				             headers=headers)

			self._logger.info("Sent tracking event {}, payload: {!r}".format(event, kwargs))
		except Exception:
			if self._logger.isEnabledFor(logging.DEBUG):
				self._logger.exception("Error while sending event to anonymous usage tracking".format(url))
			else:
				pass

__plugin_name__ = "Anonymous Usage Tracking"
__plugin_description__ = "Anonymous version and usage tracking, see homepage for details on what gets tracked"
__plugin_url__ = "https://tracking.octoprint.org"
__plugin_author__ = "Gina Häußge"

__plugin_implementation__ = TrackingPlugin()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
This module provides a bunch of utility methods and helpers FOR DEVELOPMENT ONLY.
"""

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'

import contextlib

@contextlib.contextmanager
def duration_log(context=None, log=None):
	"""
	Context manager to log the duration of the wrapped code block.

	If no ``log`` function is provided will use a ``debug`` python logger for
	``octoprint.util.dev``.

	``context`` can be set to give some textual context in the output.

	Arguments:
	    context (str): A custom string to give some textual context in the output.
	    log (callable): The log function to use to log the execution duration.
	"""
	if log is None:
		import logging
		log = logging.getLogger(__name__).debug

	import octoprint.util
	start = octoprint.util.monotonic_time()
	try:
		yield
	finally:
		end = octoprint.util.monotonic_time()
		duration = end - start

		if context:
			message = "Execution of {name} took {duration}s"
		else:
			message = "Execution of codeblock took {duration}s"

		log(message.format(name=context, duration=duration))


def log_duration(log=None, with_args=False):
	"""
	Decorator that logs the execution duration of the annotated function.

	Arguments:
	    log (callable): The logging function to use.
	    with_args (bool): Whether to include the calling arguments in the logged
	        output or not.
	"""
	def decorator(f):
		def wrapped(*args, **kwargs):
			if with_args:
				args_str = ", ".join(map(lambda x: repr(x), args))
				kwargs_str = ", ".join(map(lambda item: "{}={}".format(item[0], repr(item[1])), kwargs.items()))
				sep = ", " if args_str and kwargs_str else ""
				arguments = "".join([args_str, sep, kwargs_str])
			else:
				arguments = "..."
			context = "{name}({arguments})".format(name=f.__name__, arguments=arguments)
			with duration_log(context=context, log=log):
				return f(*args, **kwargs)
		return wrapped
	return decorator



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
This module bundles all of OctoPrint's supported plugin implementation types as well as their common parent
class, :class:`OctoPrintPlugin`.

Please note that the plugin implementation types are documented in the section
:ref:`Available plugin mixins <sec-plugins-mixins>`.

.. autoclass:: OctoPrintPlugin
   :show-inheritance:
   :members:

.. autoclass:: ReloadNeedingPlugin
   :show-inheritance:
   :members:

"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


from .core import (Plugin, RestartNeedingPlugin, SortablePlugin)

class OctoPrintPlugin(Plugin):
	"""
	The parent class of all OctoPrint plugin mixins.

	.. attribute:: _plugin_manager

	   The :class:`~octoprint.plugin.core.PluginManager` instance. Injected by the plugin core system upon
	   initialization of the implementation.

	.. attribute:: _printer_profile_manager

	   The :class:`~octoprint.printer.profile.PrinterProfileManager` instance. Injected by the plugin core system upon
	   initialization of the implementation.

	.. attribute:: _event_bus

	   The :class:`~octoprint.events.EventManager` instance. Injected by the plugin core system upon initialization of
	   the implementation.

	.. attribute:: _analysis_queue

	   The :class:`~octoprint.filemanager.analysis.AnalysisQueue` instance. Injected by the plugin core system upon
	   initialization of the implementation.

	.. attribute:: _slicing_manager

	   The :class:`~octoprint.slicing.SlicingManager` instance. Injected by the plugin core system upon initialization
	   of the implementation.

	.. attribute:: _file_manager

	   The :class:`~octoprint.filemanager.FileManager` instance. Injected by the plugin core system upon initialization
	   of the implementation.

	.. attribute:: _printer

	   The :class:`~octoprint.printer.PrinterInterface` instance. Injected by the plugin core system upon initialization
	   of the implementation.

	.. attribute:: _app_session_manager

	   The :class:`~octoprint.access.users.SessionManager` instance. Injected by the plugin core system upon initialization of
	   the implementation.

	.. attribute:: _plugin_lifecycle_manager

	   The :class:`~octoprint.server.LifecycleManager` instance. Injected by the plugin core system upon initialization
	   of the implementation.

	.. attribute:: _user_manager

	   The :class:`~octoprint.access.users.UserManager` instance. Injected by the plugin core system upon initialization
	   of the implementation.

	.. attribute:: _connectivity_checker

	   The :class:`~octoprint.util.ConnectivityChecker` instance. Injected by the plugin core system upon initialization
	   of the implementation.

	.. attribute:: _data_folder

	   Path to the data folder for the plugin to use for any data it might have to persist. Should always be accessed
	   through :meth:`get_plugin_data_folder` since that function will also ensure that the data folder actually exists
	   and if not creating it before returning it. Injected by the plugin core system upon initialization of the
	   implementation.
	"""

	# noinspection PyMissingConstructor
	def __init__(self):
		self._plugin_manager = None
		self._printer_profile_manager = None
		self._event_bus = None
		self._analysis_queue = None
		self._slicing_manager = None
		self._file_manager = None
		self._printer = None
		self._app_session_manager = None
		self._plugin_lifecycle_manager = None
		self._user_manager = None
		self._connectivity_checker = None
		self._data_folder = None

	def get_plugin_data_folder(self):
		"""
		Retrieves the path to a data folder specifically for the plugin, ensuring it exists and if not creating it
		before returning it.

		Plugins may use this folder for storing additional data they need for their operation.
		"""
		if self._data_folder is None:
			raise RuntimeError("self._plugin_data_folder is None, has the plugin been initialized yet?")

		import os
		if not os.path.isdir(self._data_folder):
			os.makedirs(self._data_folder)
		return self._data_folder


class ReloadNeedingPlugin(Plugin):
	"""
	Mixin for plugin types that need a reload of the UI after enabling/disabling them.
	"""


class EnvironmentDetectionPlugin(OctoPrintPlugin, RestartNeedingPlugin):

	def get_additional_environment(self):
		pass

	def on_environment_detected(self, environment, *args, **kwargs):
		pass


class StartupPlugin(OctoPrintPlugin, SortablePlugin):
	"""
	The ``StartupPlugin`` allows hooking into the startup of OctoPrint. It can be used to start up additional services
	on or just after the startup of the server.

	``StartupPlugin`` is a :class:`~octoprint.plugin.core.SortablePlugin` and provides
	sorting contexts for :meth:`~octoprint.plugin.StartupPlugin.on_startup` as well as
	:meth:`~octoprint.plugin.StartupPlugin.on_after_startup`.
	"""

	def on_startup(self, host, port):
		"""
		Called just before the server is actually launched. Plugins get supplied with the ``host`` and ``port`` the server
		will listen on. Note that the ``host`` may be ``0.0.0.0`` if it will listen on all interfaces, so you can't just
		blindly use this for constructing publicly reachable URLs. Also note that when this method is called, the server
		is not actually up yet and none of your plugin's APIs or blueprints will be reachable yet. If you need to be
		externally reachable, use :func:`on_after_startup` instead or additionally.

		.. warning::

		   Do not perform long-running or even blocking operations in your implementation or you **will** block and break the server.

		The relevant sorting context is ``StartupPlugin.on_startup``.

		:param string host: the host the server will listen on, may be ``0.0.0.0``
		:param int port:    the port the server will listen on
		"""

		pass

	def on_after_startup(self):
		"""
		Called just after launch of the server, so when the listen loop is actually running already.

		.. warning::

		   Do not perform long-running or even blocking operations in your implementation or you **will** block and break the server.

		The relevant sorting context is ``StartupPlugin.on_after_startup``.
		"""

		pass


class ShutdownPlugin(OctoPrintPlugin, SortablePlugin):
	"""
	The ``ShutdownPlugin`` allows hooking into the shutdown of OctoPrint. It's usually used in conjunction with the
	:class:`StartupPlugin` mixin, to cleanly shut down additional services again that where started by the :class:`StartupPlugin`
	part of the plugin.

	``ShutdownPlugin`` is a :class:`~octoprint.plugin.core.SortablePlugin` and provides a sorting context for
	:meth:`~octoprint.plugin.ShutdownPlugin.on_shutdown`.
	"""

	def on_shutdown(self):
		"""
		Called upon the imminent shutdown of OctoPrint.

		.. warning::

		   Do not perform long-running or even blocking operations in your implementation or you **will** block and break the server.

		The relevant sorting context is ``ShutdownPlugin.on_shutdown``.
		"""
		pass


class AssetPlugin(OctoPrintPlugin, RestartNeedingPlugin):
	"""
	The ``AssetPlugin`` mixin allows plugins to define additional static assets such as JavaScript or CSS files to
	be automatically embedded into the pages delivered by the server to be used within the client sided part of
	the plugin.

	A typical usage of the ``AssetPlugin`` functionality is to embed a custom view model to be used by templates injected
	through a :class:`TemplatePlugin`.

	``AssetPlugin`` is a :class:`~octoprint.plugins.core.RestartNeedingPlugin`.
	"""

	def get_asset_folder(self):
		"""
		Defines the folder where the plugin stores its static assets as defined in :func:`get_assets`. Override this if
		your plugin stores its assets at some other place than the ``static`` sub folder in the plugin base directory.

		:return string: the absolute path to the folder where the plugin stores its static assets
		"""
		import os
		return os.path.join(self._basefolder, "static")

	def get_assets(self):
		"""
		Defines the static assets the plugin offers. The following asset types are recognized and automatically
		imported at the appropriate places to be available:

		js
		   JavaScript files, such as additional view models
		jsclient
		   JavaScript files containing additional parts for the JS Client Library (since 1.3.10)
		css
		   CSS files with additional styles, will be embedded into delivered pages when not running in LESS mode.
		less
		   LESS files with additional styles, will be embedded into delivered pages when running in LESS mode.

		The expected format to be returned is a dictionary mapping one or more of these keys to a list of files of that
		type, the files being represented as relative paths from the asset folder as defined via :func:`get_asset_folder`.
		Example:

		.. code-block:: python

		   def get_assets(self):
		       return dict(
		           js=['js/my_file.js', 'js/my_other_file.js'],
		           clientjs=['clientjs/my_file.js'],
		           css=['css/my_styles.css'],
		           less=['less/my_styles.less']
		        )

		The assets will be made available by OctoPrint under the URL ``/plugin/<plugin identifier>/static/<path>``, with
		``plugin identifier`` being the plugin's identifier and ``path`` being the path as defined in the asset dictionary.

		Assets of the types ``js``, ``css`` and ``less`` will be automatically bundled by OctoPrint using
		`Flask-Assets <http://flask-assets.readthedocs.org/en/latest/>`_.

		:return dict: a dictionary describing the static assets to publish for the plugin
		"""
		return dict()


class TemplatePlugin(OctoPrintPlugin, ReloadNeedingPlugin):
	"""
	Using the ``TemplatePlugin`` mixin plugins may inject their own components into the OctoPrint web interface.

	Currently OctoPrint supports the following types of injections out of the box:

	Navbar
	   The right part of the navigation bar located at the top of the UI can be enriched with additional links. Note that
	   with the current implementation, plugins will always be located *to the left* of the existing links.

	   The included template must be called ``<plugin identifier>_navbar.jinja2`` (e.g. ``myplugin_navbar.jinja2``) unless
	   overridden by the configuration supplied through :func:`get_template_configs`.

	   The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
	   wrapper structure will have all additional classes and styles applied as specified via the configuration supplied
	   through :func:`get_template_configs`.

	Sidebar
	   The left side bar containing Connection, State and Files sections can be enriched with additional sections. Note
	   that with the current implementations, plugins will always be located *beneath* the existing sections.

	   The included template must be called ``<plugin identifier>_sidebar.jinja2`` (e.g. ``myplugin_sidebar.jinja2``) unless
	   overridden by the configuration supplied through :func:`get_template_configs`.

	   The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
	   wrapper divs for both the whole box as well as the content pane will have all additional classes and styles applied
	   as specified via the configuration supplied through :func:`get_template_configs`.

	Tabs
	   The available tabs of the main part of the interface may be extended with additional tabs originating from within
	   plugins. Note that with the current implementation, plugins will always be located *to the right* of the existing
	   tabs.

	   The included template must be called ``<plugin identifier>_tab.jinja2`` (e.g. ``myplugin_tab.jinja2``) unless
	   overridden by the configuration supplied through :func:`get_template_configs`.

	   The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
	   wrapper div and the link in the navigation will have the additional classes and styles applied as specified via the
	   configuration supplied through :func:`get_template_configs`.

	Settings
	   Plugins may inject a dialog into the existing settings view. Note that with the current implementation, plugins
	   will always be listed beneath the "Plugins" header in the settings link list, ordered alphabetically after
	   their displayed name.

	   The included template must be called ``<plugin identifier>_settings.jinja2`` (e.g. ``myplugin_settings.jinja2``) unless
	   overridden by the configuration supplied through :func:`get_template_configs`.

	   The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
	   wrapper div and the link in the navigation will have the additional classes and styles applied as defined via the
	   configuration through :func:`get_template_configs`.

	Wizards
	   Plugins may define wizard dialogs to display to the user if necessary (e.g. in case of missing information that
	   needs to be queried from the user to make the plugin work). Note that with the current implementation, all
	   wizard dialogs will be will always be sorted by their ``mandatory`` attribute (which defaults to ``False``) and then
	   alphabetically by their ``name``. Hence, mandatory wizard steps will come first, sorted alphabetically, then the
	   optional steps will follow, also alphabetically. A wizard dialog provided through a plugin will only be displayed
	   if the plugin reports the wizard as being required through :meth:`~octoprint.plugin.WizardPlugin.is_wizard_required`.
	   Please also refer to the :class:`~octoprint.plugin.WizardPlugin` mixin for further details on this.

	   The included template must be called ``<plugin identifier>_wizard.jinja2`` (e.g. ``myplugin_wizard.jinja2``) unless
	   overridden by the configuration supplied through :func:`get_template_configs`.

	   The template will be already wrapped into the necessary structure, plugins just need to supply the pure content.
	   The wrapper div and the link in the wizard navigation will have the additional classes and styles applied as defined
	   via the configuration supplied through :func:`get_template_configs`.

	   .. note::

	      A note about ``mandatory`` wizard steps: In the current implementation, marking a wizard step as
	      mandatory will *only* make it styled accordingly. It is the task of the :ref:`view model <sec-plugins-viewmodels>`
	      to actually prevent the user from skipping the dialog by implementing the ``onWizardTabChange``
	      callback and returning ``false`` there if it is detected that the user hasn't yet filled in the
	      wizard step.

	About
	   Plugins may define additional panels into OctoPrint's "About" dialog. Note that with the current implementation
	   further about dialog panels will be sorted alphabetically by their name and sorted after the predefined ones.

	   The included template must be called ``<plugin identifier>_about.jinja2`` (e.g. ``myplugin_about.jinja2``) unless
	   overridden by the configuration supplied through :func:`get_template_configs`.

	   The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
	   wrapped div and the link in the navigation will have the additional classes and styles applied as defined via
	   the configuration supplied through :func:`get_template_configs`.

	Generic
	   Plugins may also inject arbitrary templates into the page of the web interface itself, e.g. in order to
	   add overlays or dialogs to be called from within the plugin's JavaScript code.

	.. figure:: ../images/template-plugin-types-main.png
	   :align: center
	   :alt: Template injection types in the main part of the interface

	   Template injection types in the main part of the interface

	.. figure:: ../images/template-plugin-types-settings.png
	   :align: center
	   :alt: Template injection types in the settings

	   Template injection types in the settings

	You can find an example for a simple plugin which injects navbar, tab and settings content into the interface in
	the "helloworld" plugin in OctoPrint's :ref:`Plugin Tutorial <sec-plugins-gettingstarted>`.

	Plugins may replace existing components, see the ``replaces`` keyword in the template configurations returned by
	:meth:`.get_template_configs` below. Note that if a plugin replaces a core component, it is the plugin's
	responsibility to ensure that all core functionality is still maintained.

	Plugins can also add additional template types by implementing the :ref:`octoprint.ui.web.templatetypes <sec-plugins-hook-ui-web-templatetypes>` hook.

	``TemplatePlugin`` is a :class:`~octoprint.plugin.core.ReloadNeedingPlugin`.
	"""

	def get_template_configs(self):
		"""
		Allows configuration of injected navbar, sidebar, tab and settings templates (and also additional templates of
		types specified by plugins through the :ref:`octoprint.ui.web.templatetypes <sec-plugins-hook-ui-web-templatetypes>` hook).
		Should be a list containing one configuration object per template to inject. Each configuration object is
		represented by a dictionary which may contain the following keys:

		.. list-table::
		   :widths: 5 95

		   * - type
		     - The template type the configuration is targeting. Possible values here are ``navbar``, ``sidebar``,
		       ``tab``, ``settings`` and ``generic``. Mandatory.
		   * - name
		     - The name of the component, if not set the name of the plugin will be used. The name will be visible at
		       a location depending on the ``type``:

		         * ``navbar``: unused
		         * ``sidebar``: sidebar heading
		         * ``tab``: tab heading
		         * ``settings``: settings link
		         * ``wizard``: wizard link
		         * ``about``: about link
		         * ``generic``: unused

		   * - template
		     - Name of the template to inject, default value depends on the ``type``:

		         * ``navbar``: ``<plugin identifier>_navbar.jinja2``
		         * ``sidebar``: ``<plugin identifier>_sidebar.jinja2``
		         * ``tab``: ``<plugin identifier>_tab.jinja2``
		         * ``settings``: ``<plugin identifier>_settings.jinja2``
		         * ``wizard``: ``<plugin identifier>_wizard.jinja2``
		         * ``about``: ``<plugin identifier>_about.jinja2``
		         * ``generic``: ``<plugin identifier>.jinja2``

		   * - suffix
		     - Suffix to attach to the component identifier and the div identifier of the injected template. Will be
		       ``_<index>`` if not provided and not the first template of the type, with ``index`` counting from 1 and
		       increasing for each template of the same type.

		       Example: If your plugin with identifier ``myplugin`` defines two tab components like this:

		       .. code-block:: python

		          return [
		              dict(type="tab", template="myplugin_first_tab.jinja2"),
		              dict(type="tab", template="myplugin_second_tab.jinja2")
		          ]

		       then the first tab will have the component identifier ``plugin_myplugin`` and the second one will have
		       the component identifier ``plugin_myplugin_2`` (the generated divs will be ``tab_plugin_myplugin`` and
		       ``tab_plugin_myplugin_2`` accordingly). Notice that the first tab is *not* called ``plugin_myplugin_1`` --
		       as stated above while the ``index`` used as default suffix starts counting at 1, it will not be applied
		       for the first component of a given type.

		       If on the other hand your plugin's definition looks like this:

		       .. code-block:: python

		          return [
		              dict(type="tab", template="myplugin_first_tab_jinja2", suffix="_1st"),
		              dict(type="tab", template="myplugin_second_tab_jinja2", suffix="_2nd")
		          ]

		       then the generated component identifier will be ``plugin_myplugin_1st`` and ``plugin_myplugin_2nd``
		       (and the divs will be ``tab_plugin_myplugin_1st`` and ``tab_plugin_myplugin_2nd``).

		   * - div
		     - Id for the div containing the component. If not provided, defaults to ``<type>_plugin_<plugin identifier>`` plus
		       the ``suffix`` if provided or required.
		   * - replaces
		     - Id of the component this one replaces, might be either one of the core components or a component
		       provided by another plugin. A list of the core component identifiers can be found
		       :ref:`in the configuration documentation <sec-configuration-config_yaml-appearance>`. The identifiers of
		       other plugin components always follow the format described above.
		   * - custom_bindings
		     - A boolean value indicating whether the default view model should be bound to the component (``false``)
		       or if a custom binding will be used by the plugin (``true``, default).
		   * - data_bind
		     - Additional knockout data bindings to apply to the component, can be used to add further behaviour to
		       the container based on internal state if necessary.
		   * - classes
		     - Additional classes to apply to the component, as a list of individual classes
		       (e.g. ``classes=["myclass", "myotherclass"]``) which will be joined into the correct format by the template engine.
		   * - styles
		     - Additional CSS styles to apply to the component, as a list of individual declarations
		       (e.g. ``styles=["color: red", "display: block"]``) which will be joined into the correct format by the template
		       engine.

		Further keys to be included in the dictionary depend on the type:

		``sidebar`` type

		   .. list-table::
		      :widths: 5 95

		      * - icon
		        - Icon to use for the sidebar header, should be the name of a Font Awesome icon without the leading ``icon-`` part.
		      * - template_header
		        - Additional template to include in the head section of the sidebar item. For an example of this, see the additional
		          options included in the "Files" section.
		      * - classes_wrapper
		        - Like ``classes`` but only applied to the whole wrapper around the sidebar box.
		      * - classes_content
		        - Like ``classes`` but only applied to the content pane itself.
		      * - styles_wrapper
		        - Like ``styles`` but only applied to the whole wrapper around the sidebar box.
		      * - styles_content
		        - Like ``styles`` but only applied to the content pane itself

		``tab`` type and ``settings`` type

		   .. list-table::
		      :widths: 5 95

		      * - classes_link
		        - Like ``classes`` but only applied to the link in the navigation.
		      * - classes_content
		        - Like ``classes`` but only applied to the content pane itself.
		      * - styles_link
		        - Like ``styles`` but only applied to the link in the navigation.
		      * - styles_content
		        - Like ``styles`` but only applied to the content pane itself.

		``wizard`` type

		   .. list-table::
		      :widths: 5 95

		      * - mandatory
		        - Whether the wizard step is mandatory (True) or not (False). Optional,
		          defaults to False. If set to True, OctoPrint will sort visually mark
		          the step as mandatory in the UI (bold in the navigation and a little
		          alert) and also sort it into the first half.

		.. note::

		   As already outlined above, each template type has a default template name (i.e. the default navbar template
		   of a plugin is called ``<plugin identifier>_navbar.jinja2``), which may be overridden using the template configuration.
		   If a plugin needs to include more than one template of a given type, it needs to provide an entry for each of
		   those, since the implicit default template will only be included automatically if no other templates of that
		   type are defined.

		   Example: If you have a plugin that injects two tab components, one defined in the template file
		   ``myplugin_tab.jinja2`` (the default template) and one in the template ``myplugin_othertab.jinja2``, you
		   might be tempted to just return the following configuration since one your templates is named by the default
		   template name:

		   .. code-block:: python

		      return [
		          dict(type="tab", template="myplugin_othertab.jinja2")
		      ]

		   This will only include the tab defined in ``myplugin_othertab.jinja2`` though, ``myplugin_tab.jinja2`` will
		   not be included automatically since the presence of a definition for the ``tab`` type overrides the automatic
		   injection of the default template. You'll have to include it explicitly:

		   .. code-block:: python

		      return [
		          dict(type="tab", template="myplugin_tab.jinja2"),
		          dict(type="tab", template="myplugin_othertab.jinja2")
		      ]

		:return list: a list containing the configuration options for the plugin's injected templates
		"""
		return []

	def get_template_vars(self):
		"""
		Defines additional template variables to include into the template renderer. Variable names will be prefixed
		with ``plugin_<plugin identifier>_``.

		:return dict: a dictionary containing any additional template variables to include in the renderer
		"""
		return dict()

	def get_template_folder(self):
		"""
		Defines the folder where the plugin stores its templates. Override this if your plugin stores its templates at
		some other place than the ``templates`` sub folder in the plugin base directory.

		:return string: the absolute path to the folder where the plugin stores its jinja2 templates
		"""
		import os
		return os.path.join(self._basefolder, "templates")


class UiPlugin(OctoPrintPlugin, SortablePlugin):
	"""
	The ``UiPlugin`` mixin allows plugins to completely replace the UI served
	by OctoPrint when requesting the main page hosted at `/`.

	OctoPrint will query whether your mixin implementation will handle a
	provided request by calling :meth:`~octoprint.plugin.UiPlugin.will_handle_ui` with the Flask
	`Request <http://flask.pocoo.org/docs/0.10/api/#flask.Request>`_ object as
	parameter. If you plugin returns `True` here, OctoPrint will next call
	:meth:`~octoprint.plugin.UiPlugin.on_ui_render` with a few parameters like
	- again - the Flask Request object and the render keyword arguments as
	used by the default OctoPrint web interface. For more information see below.

	There are two methods used in order to allow for caching of the actual
	response sent to the client. Whatever a plugin implementation returns
	from the call to its :meth:`~octoprint.plugin.UiPlugin.on_ui_render` method
	will be cached server side. The cache will be emptied in case of explicit
	no-cache headers sent by the client, or if the ``_refresh`` query parameter
	on the request exists and is set to ``true``. To prevent caching of the
	response altogether, a plugin may set no-cache headers on the returned
	response as well.

	``UiPlugin`` is a :class:`~octoprint.plugin.core.SortablePlugin` with a sorting context
	for :meth:`~octoprint.plugin.UiPlugin.will_handle_ui`. The first plugin to return ``True``
	for :meth:`~octoprint.plugin.UiPlugin.will_handle_ui` will be the one whose ui will be used,
	no further calls to :meth:`~octoprint.plugin.UiPlugin.on_ui_render` will be performed.

	If implementations want to serve custom templates in the :meth:`~octoprint.plugin.UiPlugin.on_ui_render`
	method it is recommended to also implement the :class:`~octoprint.plugin.TemplatePlugin`
	mixin.

	**Example**

	What follows is a very simple example that renders a different (non functional and
	only exemplary) UI if the requesting client has a UserAgent string hinting
	at it being a mobile device:

	.. onlineinclude:: https://raw.githubusercontent.com/OctoPrint/Plugin-Examples/master/dummy_mobile_ui/__init__.py
	   :linenos:
	   :tab-width: 4
	   :caption: `dummy_mobile_ui/__init__.py <https://github.com/OctoPrint/Plugin-Examples/blob/master/dummy_mobile_ui/__init__.py>`_

	.. onlineinclude:: https://raw.githubusercontent.com/OctoPrint/Plugin-Examples/master/dummy_mobile_ui/templates/dummy_mobile_ui_index.jinja2
	   :linenos:
	   :tab-width: 4
	   :caption: `dummy_mobile_ui/templates/dummy_mobile_ui_index.jinja2 <https://github.com/OctoPrint/Plugin-Examples/blob/master/dummy_mobile_ui/templates/dummy_mobile_ui_index.jinja2>`_

	Try installing the above plugin ``dummy_mobile_ui`` (also available in the
	`plugin examples repository <https://github.com/OctoPrint/Plugin-Examples/blob/master/dummy_mobile_ui>`_)
	into your OctoPrint instance. If you access it from a regular desktop browser,
	you should still see the default UI. However if you access it from a mobile
	device (make sure to not have that request the desktop version of pages!)
	you should see the very simple dummy page defined above.

	**Preemptive and Runtime Caching**

	OctoPrint will also cache your custom UI for you in its server side UI cache, making sure
	it only gets re-rendered if the request demands that (by having no-cache headers set) or if
	the cache gets invalidated otherwise.

	In order to be able to do that, the ``UiPlugin`` offers overriding some cache specific
	methods used for figuring out the source files whose modification time to use for cache invalidation
	as well as override possibilities for ETag and LastModified calculation. Additionally there are
	methods to allow persisting call parameters to allow for preemptively caching your UI during
	server startup (basically eager caching instead of lazily waiting for the first request).

	See below for details on this.
	"""

	# noinspection PyMethodMayBeStatic,PyUnusedLocal
	def will_handle_ui(self, request):
		"""
		Called by OctoPrint to determine if the mixin implementation will be
		able to handle the ``request`` provided as a parameter.

		Return ``True`` here to signal that your implementation will handle
		the request and that the result of its :meth:`~octoprint.plugin.UiPlugin.on_ui_render` method
		is what should be served to the user.

		The execution order of calls to this method can be influenced via the sorting context
		``UiPlugin.will_handle_ui``.

		Arguments:
		    request (flask.Request): A Flask `Request <http://flask.pocoo.org/docs/0.10/api/#flask.Request>`_
		        object.

		Returns:
		    bool: ``True`` if the the implementation will serve the request,
		        ``False`` otherwise.
		"""
		return False

	# noinspection PyMethodMayBeStatic,PyUnusedLocal
	def on_ui_render(self, now, request, render_kwargs):
		"""
		Called by OctoPrint to retrieve the response to send to the client
		for the ``request`` to ``/``. Only called if :meth:`~octoprint.plugin.UiPlugin.will_handle_ui`
		returned ``True``.

		``render_kwargs`` will be a dictionary (whose contents are cached) which
		will contain the following key and value pairs (note that not all
		key value pairs contained in the dictionary are listed here, only
		those you should depend on as a plugin developer at the current time):

		.. list-table::
		   :widths: 5 95

		   * - debug
		     - ``True`` if debug mode is enabled, ``False`` otherwise.
		   * - firstRun
		     - ``True`` if the server is being run for the first time (not
		       configured yet), ``False`` otherwise.
		   * - version
		     - OctoPrint's version information. This is a ``dict`` with the
		       following keys:

		       .. list-table::
		          :widths: 5 95

		          * - number
		            - The version number (e.g. ``x.y.z``)
		          * - branch
		            - The GIT branch from which the OctoPrint instance was built
		              (e.g. ``master``)
		          * - display
		            - The full human readable version string, including the
		              branch information (e.g. ``x.y.z (master branch)``

		   * - uiApiKey
		     - The UI API key to use for unauthorized API requests. This is
		       freshly generated on every server restart.
		   * - templates
		     - Template data to render in the UI. Will be a ``dict`` containing entries
		       for all known template types.

		       The sub structure for each key will be as follows:

		       .. list-table::
		          :widths: 5 95

		          * - order
		            - A list of template names in the order they should appear
		              in the final rendered page
		          * - entries
		            - The template entry definitions to render. Depending on the
		              template type those are either 2-tuples of a name and a ``dict``
		              or directly ``dicts`` with information regarding the
		              template to render.

		              For the possible contents of the data ``dicts`` see the
		              :class:`~octoprint.plugin.TemplatePlugin` mixin.

		   * - pluginNames
		     - A list of names of :class:`~octoprint.plugin.TemplatePlugin`
		       implementation that were enabled when creating the ``templates``
		       value.
		   * - locales
		     - The locales for which there are translations available.
		   * - supportedExtensions
		     - The file extensions supported for uploads.

		On top of that all additional template variables as provided by :meth:`~octoprint.plugin.TemplatePlugin.get_template_vars`
		will be contained in the dictionary as well.

		Arguments:
		    now (datetime.datetime): The datetime instance representing "now"
		        for this request, in case your plugin implementation needs this
		        information.
		    request (flask.Request): A Flask `Request <http://flask.pocoo.org/docs/0.10/api/#flask.Request>`_ object.
		    render_kwargs (dict): The (cached) render keyword arguments that
		        would usually be provided to the core UI render function.

		Returns:
		    flask.Response: Should return a Flask `Response <http://flask.pocoo.org/docs/0.10/api/#flask.Response>`_
		        object that can be served to the requesting client directly. May be
		        created with ``flask.make_response`` combined with something like
		        ``flask.render_template``.
		"""

		return None

	# noinspection PyMethodMayBeStatic
	def get_ui_additional_key_data_for_cache(self):
		"""
		Allows to return additional data to use in the cache key.

		Returns:
		    list, tuple: A list or tuple of strings to use in the cache key. Will be joined by OctoPrint
		        using ``:`` as separator and appended to the existing ``ui:<identifier>:<base url>:<locale>``
		        cache key. Ignored if ``None`` is returned.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_ui_additional_tracked_files(self):
		"""
		Allows to return additional files to track for validating existing caches. By default OctoPrint
		will track all declared templates, assets and translation files in the system. Additional
		files can be added by a plugin through this callback.

		Returns:
		    list: A list of paths to additional files whose modification to track for (in)validating
		        the cache. Ignored if ``None`` is returned.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_ui_custom_tracked_files(self):
		"""
		Allows to define a complete separate set of files to track for (in)validating the cache. If this
		method returns something, the templates, assets and translation files won't be tracked, only the
		files specified in the returned list.

		Returns:
		    list: A list of paths representing the only files whose modification to track for (in)validating
		        the cache. Ignored if ``None`` is returned.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_ui_custom_etag(self):
		"""
		Allows to use a custom way to calculate the ETag, instead of the default method (hashing
		OctoPrint's version, tracked file paths and ``LastModified`` value).

		Returns:
		    str: An alternatively calculated ETag value. Ignored if ``None`` is returned (default).
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_ui_additional_etag(self, default_additional):
		"""
		Allows to provide a list of additional fields to use for ETag generation.

		By default the same list will be returned that is also used in the stock UI (and injected
		via the parameter ``default_additional``).

		Arguments:
		    default_additional (list): The list of default fields added to the ETag of the default UI

		Returns:
		    (list): A list of additional fields for the ETag generation, or None
		"""
		return default_additional

	# noinspection PyMethodMayBeStatic
	def get_ui_custom_lastmodified(self):
		"""
		Allows to calculate the LastModified differently than using the most recent modification
		date of all tracked files.

		Returns:
		    int: An alternatively calculated LastModified value. Ignored if ``None`` is returned (default).
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_ui_preemptive_caching_enabled(self):
		"""
		Allows to control whether the view provided by the plugin should be preemptively
		cached on server startup (default) or not.

		Have this return False if you do not want your plugin's UI to ever be preemptively cached.

		Returns:
		    bool: Whether to enable preemptive caching (True, default) or not (False)
		"""
		return True

	# noinspection PyMethodMayBeStatic
	def get_ui_data_for_preemptive_caching(self):
		"""
		Allows defining additional data to be persisted in the preemptive cache configuration, on
		top of the request path, base URL and used locale.

		Returns:
		    dict: Additional data to persist in the preemptive cache configuration.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_ui_additional_request_data_for_preemptive_caching(self):
		"""
		Allows defining additional request data to persist in the preemptive cache configuration and
		to use for the fake request used for populating the preemptive cache.

		Keys and values are used as keyword arguments for creating the
		`Werkzeug EnvironBuilder <http://werkzeug.pocoo.org/docs/0.11/test/#werkzeug.test.EnvironBuilder>`_
		used for creating the fake request.

		Returns:
		    dict: Additional request data to persist in the preemptive cache configuration and to
		        use for request environment construction.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_ui_preemptive_caching_additional_unless(self):
		"""
		Allows defining additional reasons for temporarily not adding a preemptive cache record for
		your plugin's UI.

		OctoPrint will call this method when processing a UI request, to determine whether to record the
		access or not. If you return ``True`` here, no record will be created.

		Returns:
		    bool: Whether to suppress a record (True) or not (False, default)
		"""
		return False

	# noinspection PyMethodMayBeStatic
	def get_ui_custom_template_filter(self, default_template_filter):
		"""
		Allows to specify a custom template filter to use for filtering the template contained in the
		``render_kwargs`` provided to the templating sub system.

		Only relevant for UiPlugins that actually utilize the stock templates of OctoPrint.

		By default simply returns the provided ``default_template_filter``.

		Arguments:
		    default_template_filter (callable): The default template filter used by the default UI

		Returns:
		    (callable) A filter function accepting the ``template_type`` and ``template_key`` of a template
		    and returning ``True`` to keep it and ``False`` to filter it out. If ``None`` is returned, no
		    filtering will take place.
		"""
		return default_template_filter

class WizardPlugin(OctoPrintPlugin, ReloadNeedingPlugin):
	"""
	The ``WizardPlugin`` mixin allows plugins to report to OctoPrint whether
	the ``wizard`` templates they define via the :class:`~octoprint.plugin.TemplatePlugin`
	should be displayed to the user, what details to provide to their respective
	wizard frontend components and what to do when the wizard is finished
	by the user.

	OctoPrint will only display such wizard dialogs to the user which belong
	to plugins that

	  * report ``True`` in their :func:`is_wizard_required` method and
	  * have not yet been shown to the user in the version currently being reported
	    by the :meth:`~octoprint.plugin.WizardPlugin.get_wizard_version` method

	Example: If a plugin with the identifier ``myplugin`` has a specific
	setting ``some_key`` it needs to have filled by the user in order to be
	able to work at all, it would probably test for that setting's value in
	the :meth:`~octoprint.plugin.WizardPlugin.is_wizard_required` method and
	return ``True`` if the value is unset:

	.. code-block:: python

	   class MyPlugin(octoprint.plugin.SettingsPlugin,
	                  octoprint.plugin.TemplatePlugin,
	                  octoprint.plugin.WizardPlugin):

	       def get_default_settings(self):
	           return dict(some_key=None)

	       def is_wizard_required(self):
	           return self._settings.get(["some_key"]) is None

	OctoPrint will then display the wizard dialog provided by the plugin through
	the :class:`TemplatePlugin` mixin. Once the user finishes the wizard on the
	frontend, OctoPrint will store that it already showed the wizard of ``myplugin``
	in the version reported by :meth:`~octoprint.plugin.WizardPlugin.get_wizard_version`
	- here ``None`` since that is the default value returned by that function
	and the plugin did not override it.

	If the plugin in a later version needs another setting from the user in order
	to function, it will also need to change the reported version in order to
	have OctoPrint reshow the dialog. E.g.

	.. code-block:: python

	   class MyPlugin(octoprint.plugin.SettingsPlugin,
	                  octoprint.plugin.TemplatePlugin,
	                  octoprint.plugin.WizardPlugin):

	       def get_default_settings(self):
	           return dict(some_key=None, some_other_key=None)

	       def is_wizard_required(self):
	           some_key_unset = self._settings.get(["some_key"]) is None
	           some_other_key_unset = self._settings.get(["some_other_key"]) is None

	           return some_key_unset or some_other_key_unset

	       def get_wizard_version(self):
	           return 1

	``WizardPlugin`` is a :class:`~octoprint.plugin.core.ReloadNeedingPlugin`.
	"""

	# noinspection PyMethodMayBeStatic
	def is_wizard_required(self):
		"""
		Allows the plugin to report whether it needs to display a wizard to the
		user or not.

		Defaults to ``False``.

		OctoPrint will only include those wizards from plugins which are reporting
		their wizards as being required through this method returning ``True``.
		Still, if OctoPrint already displayed that wizard in the same version
		to the user once it won't be displayed again regardless whether this
		method returns ``True`` or not.
		"""
		return False

	# noinspection PyMethodMayBeStatic
	def get_wizard_version(self):
		"""
		The version of this plugin's wizard. OctoPrint will only display a wizard
		of the same plugin and wizard version once to the user. After they
		finish the wizard, OctoPrint will remember that it already showed this
		wizard in this particular version and not reshow it.

		If a plugin needs to show its wizard to the user again (e.g. because
		of changes in the required settings), increasing this value is the
		way to notify OctoPrint of these changes.

		Returns:
		    int or None: an int signifying the current wizard version, should be incremented by plugins whenever there
		                 are changes to the plugin that might necessitate reshowing the wizard if it is required. ``None``
		                 will also be accepted and lead to the wizard always be ignored unless it has never been finished
		                 so far
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_wizard_details(self):
		"""
		Called by OctoPrint when the wizard wrapper dialog is shown. Allows the plugin to return data
		that will then be made available to the view models via the view model callback ``onWizardDetails``.

		Use this if your plugin's view model that handles your wizard dialog needs additional
		data to perform its task.

		Returns:
		    dict: a dictionary containing additional data to provide to the frontend. Whatever the plugin
		          returns here will be made available on the wizard API under the plugin's identifier
		"""
		return dict()

	# noinspection PyMethodMayBeStatic,PyUnusedLocal
	def on_wizard_finish(self, handled):
		"""
		Called by OctoPrint whenever the user finishes a wizard session.

		The ``handled`` parameter will indicate whether that plugin's wizard was
		included in the wizard dialog presented to the user (so the plugin providing
		it was reporting that the wizard was required and the wizard plus version was not
		ignored/had already been seen).

		Use this to do any clean up tasks necessary after wizard completion.

		Arguments:
		    handled (bool): True if the plugin's wizard was previously reported as
		                    required, not ignored and thus presented to the user,
		                    False otherwise
		"""
		pass

	# noinspection PyProtectedMember
	@classmethod
	def is_wizard_ignored(cls, seen_wizards, implementation):
		"""
		Determines whether the provided implementation is ignored based on the
		provided information about already seen wizards and their versions or not.

		A wizard is ignored if

		  * the current and seen versions are identical
		  * the current version is None and the seen version is not
		  * the current version is less or equal than the seen one

		.. code-block:: none

		       |  current  |
		       | N | 1 | 2 |   N = None
		   ----+---+---+---+   X = ignored
		   s N | X |   |   |
		   e --+---+---+---+
		   e 1 | X | X |   |
		   n --+---+---+---+
		     2 | X | X | X |
		   ----+---+---+---+

		Arguments:
		    seen_wizards (dict): A dictionary with information about already seen
		        wizards and their versions. Mappings from the identifiers of
		        the plugin providing the wizard to the reported wizard
		        version (int or None) that was already seen by the user.
		    implementation (object): The plugin implementation to check.

		Returns:
		    bool: False if the provided ``implementation`` is either not a :class:`WizardPlugin`
		          or has not yet been seen (in this version), True otherwise
		"""

		if not isinstance(implementation, cls):
			return False

		name = implementation._identifier
		if not name in seen_wizards:
			return False

		seen = seen_wizards[name]
		wizard_version = implementation.get_wizard_version()

		current = None
		if wizard_version is not None:
			try:
				current = int(wizard_version)
			except ValueError as e:
				import logging
				logging.getLogger(__name__).log("WizardPlugin {} returned invalid value {} for wizard version: {}".format(name, wizard_version, str(e)))

		return (current == seen) \
		       or (current is None and seen is not None) \
		       or (current <= seen)


class SimpleApiPlugin(OctoPrintPlugin):
	"""
	Utilizing the ``SimpleApiPlugin`` mixin plugins may implement a simple API based around one GET resource and one
	resource accepting JSON commands POSTed to it. This is the easy alternative for plugin's which don't need the
	full power of a `Flask Blueprint <http://flask.pocoo.org/docs/0.10/blueprints/>`_ that the :class:`BlueprintPlugin`
	mixin offers.

	Use this mixin if all you need to do is return some kind of dynamic data to your plugin from the backend
	and/or want to react to simple commands which boil down to a type of command and a few flat parameters
	supplied with it.

	The simple API constructed by OctoPrint for you will be made available under ``/api/plugin/<plugin identifier>/``.
	OctoPrint will do some preliminary request validation for your defined commands, making sure the request body is in
	the correct format (content type must be JSON) and contains all obligatory parameters for your command.

	Let's take a look at a small example for such a simple API and how you would go about calling it.

	Take this example of a plugin registered under plugin identifier ``mysimpleapiplugin``:

	.. code-block:: python
	   :linenos:

	   import octoprint.plugin

	   import flask

	   class MySimpleApiPlugin(octoprint.plugin.SimpleApiPlugin):
	       def get_api_commands(self):
	           return dict(
	               command1=[],
	               command2=["some_parameter"]
	           )

	       def on_api_command(self, command, data):
	           import flask
	           if command == "command1":
	               parameter = "unset"
	               if "parameter" in data:
	                   parameter = "set"
	               self._logger.info("command1 called, parameter is {parameter}".format(**locals()))
	           elif command == "command2":
	               self._logger.info("command2 called, some_parameter is {some_parameter}".format(**data))

	       def on_api_get(self, request):
	           return flask.jsonify(foo="bar")

	   __plugin_implementation__ = MySimpleApiPlugin()

	Our plugin defines two commands, ``command1`` with no mandatory parameters and ``command2`` with one
	mandatory parameter ``some_parameter``.

	``command1`` can also accept an optional parameter ``parameter``, and will log whether
	that parameter was set or unset. ``command2`` will log the content of the mandatory ``some_parameter`` parameter.

	A valid POST request for ``command2`` sent to ``/api/plugin/mysimpleapiplugin`` would look like this:

	.. sourcecode:: http

	   POST /api/plugin/mysimpleapiplugin HTTP/1.1
	   Host: example.com
	   Content-Type: application/json
	   X-Api-Key: abcdef...

	   {
	     "command": "command2",
	     "some_parameter": "some_value",
	     "some_optional_parameter": 2342
	   }

	which would produce a response like this:

	.. sourcecode:: http

	   HTTP/1.1 204 No Content

	and print something like this line to ``octoprint.log``::

	   2015-02-12 17:40:21,140 - octoprint.plugins.mysimpleapiplugin - INFO - command2 called, some_parameter is some_value

	A GET request on our plugin's simple API resource will only return a JSON document like this:

	.. sourcecode:: http

	   HTTP/1.1 200 Ok
	   Content-Type: application/json

	   {
	     "foo": "bar"
	   }
	"""

	# noinspection PyMethodMayBeStatic
	def get_api_commands(self):
		"""
		Return a dictionary here with the keys representing the accepted commands and the values being lists of
		mandatory parameter names.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def is_api_adminonly(self):
		"""
		Return True if the API is only available to users having the admin role.
		"""
		return False

	# noinspection PyMethodMayBeStatic
	def on_api_command(self, command, data):
		"""
		Called by OctoPrint upon a POST request to ``/api/plugin/<plugin identifier>``. ``command`` will contain one of
		the commands as specified via :func:`get_api_commands`, ``data`` will contain the full request body parsed
		from JSON into a Python dictionary. Note that this will also contain the ``command`` attribute itself. For the
		example given above, for the ``command2`` request the ``data`` received by the plugin would be equal to
		``dict(command="command2", some_parameter="some_value")``.

		If your plugin returns nothing here, OctoPrint will return an empty response with return code ``204 No content``
		for you. You may also return regular responses as you would return from any Flask view here though, e.g.
		``return flask.jsonify(result="some json result")`` or ``return flask.make_response("Not found", 404)``.

		:param string command: the command with which the resource was called
		:param dict data:      the full request body of the POST request parsed from JSON into a Python dictionary
		:return: ``None`` in which case OctoPrint will generate a ``204 No content`` response with empty body, or optionally
		         a proper Flask response.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def on_api_get(self, request):
		"""
		Called by OctoPrint upon a GET request to ``/api/plugin/<plugin identifier>``. ``request`` will contain the
		received `Flask request object <http://flask.pocoo.org/docs/0.9/api/#flask.Request>`_ which you may evaluate
		for additional arguments supplied with the request.

		If your plugin returns nothing here, OctoPrint will return an empty response with return code ``204 No content``
		for you. You may also return regular responses as you would return from any Flask view here though, e.g.
		``return flask.jsonify(result="some json result")`` or ``return flask.make_response("Not found", 404)``.

		:param request: the Flask request object
		:return: ``None`` in which case OctoPrint will generate a ``204 No content`` response with empty body, or optionally
		         a proper Flask response.
		"""
		return None


class BlueprintPlugin(OctoPrintPlugin, RestartNeedingPlugin):
	"""
	The ``BlueprintPlugin`` mixin allows plugins to define their own full fledged endpoints for whatever purpose,
	be it a more sophisticated API than what is possible via the :class:`SimpleApiPlugin` or a custom web frontend.

	The mechanism at work here is `Flask's <http://flask.pocoo.org/>`_ own `Blueprint mechanism <http://flask.pocoo.org/docs/0.10/blueprints/>`_.

	The mixin automatically creates a blueprint for you that will be registered under ``/plugin/<plugin identifier>/``.
	All you need to do is decorate all of your view functions with the :func:`route` decorator,
	which behaves exactly the same like Flask's regular ``route`` decorators. Example:

	.. code-block:: python
	   :linenos:

	   import octoprint.plugin
	   import flask

	   class MyBlueprintPlugin(octoprint.plugin.BlueprintPlugin):
	       @octoprint.plugin.BlueprintPlugin.route("/echo", methods=["GET"])
	       def myEcho(self):
	           if not "text" in flask.request.values:
	               return flask.make_response("Expected a text to echo back.", 400)
	           return flask.request.values["text"]

	   __plugin_implementation__ = MyBlueprintPlugin()

	Your blueprint will be published by OctoPrint under the base URL ``/plugin/<plugin identifier>/``, so the above
	example of a plugin with the identifier "myblueprintplugin" would be reachable under
	``/plugin/myblueprintplugin/echo``.

	Just like with regular blueprints you'll be able to create URLs via ``url_for``, just use the prefix
	``plugin.<plugin identifier>.<method_name>``, e.g.:

	.. code-block:: python

	   flask.url_for("plugin.myblueprintplugin.myEcho") # will return "/plugin/myblueprintplugin/echo"


	``BlueprintPlugin`` implements :class:`~octoprint.plugins.core.RestartNeedingPlugin`.
	"""

	@staticmethod
	def route(rule, **options):
		"""
		A decorator to mark view methods in your BlueprintPlugin subclass. Works just the same as Flask's
		own ``route`` decorator available on blueprints.

		See `the documentation for flask.Blueprint.route <http://flask.pocoo.org/docs/0.10/api/#flask.Blueprint.route>`_
		and `the documentation for flask.Flask.route <http://flask.pocoo.org/docs/0.10/api/#flask.Flask.route>`_ for more
		information.
		"""

		from collections import defaultdict
		def decorator(f):
			# We attach the decorator parameters directly to the function object, because that's the only place
			# we can access right now.
			# This neat little trick was adapter from the Flask-Classy project: https://pythonhosted.org/Flask-Classy/
			if not hasattr(f, "_blueprint_rules") or f._blueprint_rules is None:
				f._blueprint_rules = defaultdict(list)
			f._blueprint_rules[f.__name__].append((rule, options))
			return f
		return decorator

	@staticmethod
	def errorhandler(code_or_exception):
		"""
		A decorator to mark errorhandlings methods in your BlueprintPlugin subclass. Works just the same as Flask's
		own ``errorhandler`` decorator available on blueprints.

		See `the documentation for flask.Blueprint.errorhandler <http://flask.pocoo.org/docs/0.10/api/#flask.Blueprint.errorhandler>`_
		and `the documentation for flask.Flask.errorhandler <http://flask.pocoo.org/docs/0.10/api/#flask.Flask.errorhandler>`_ for more
		information.
		"""
		from collections import defaultdict
		def decorator(f):
			if not hasattr(f, "_blueprint_error_handler") or f._blueprint_error_handler is None:
				f._blueprint_error_handler = defaultdict(list)
			f._blueprint_error_handler[f.__name__].append(code_or_exception)
			return f
		return decorator

	# noinspection PyProtectedMember
	def get_blueprint(self):
		"""
		Creates and returns the blueprint for your plugin. Override this if you want to define and handle your blueprint yourself.

		This method will only be called once during server initialization.

		:return: the blueprint ready to be registered with Flask
		"""

		if hasattr(self, "_blueprint"):
			# if we already constructed the blueprint and hence have it cached,
			# return that instance - we don't want to instance it multiple times
			return self._blueprint

		import flask
		kwargs = self.get_blueprint_kwargs()
		blueprint = flask.Blueprint("plugin." + self._identifier, self._identifier, **kwargs)

		# we now iterate over all members of ourselves and look if we find an attribute
		# that has data originating from one of our decorators - we ignore anything
		# starting with a _ to only handle public stuff
		for member in [member for member in dir(self) if not member.startswith("_")]:
			f = getattr(self, member)

			if hasattr(f, "_blueprint_rules") and member in f._blueprint_rules:
				# this attribute was annotated with our @route decorator
				for blueprint_rule in f._blueprint_rules[member]:
					rule, options = blueprint_rule
					blueprint.add_url_rule(rule, options.pop("endpoint", f.__name__), view_func=f, **options)

			if hasattr(f, "_blueprint_error_handler") and member in f._blueprint_error_handler:
				# this attribute was annotated with our @error_handler decorator
				for code_or_exception in f._blueprint_error_handler[member]:
					blueprint.errorhandler(code_or_exception)(f)

		# cache and return the blueprint object
		self._blueprint = blueprint
		return blueprint

	def get_blueprint_kwargs(self):
		"""
		Override this if you want your blueprint constructed with additional options such as ``static_folder``,
		``template_folder``, etc.

		Defaults to the blueprint's ``static_folder`` and ``template_folder`` to be set to the plugin's basefolder
		plus ``/static`` or respectively ``/templates``, or -- if the plugin also implements :class:`AssetPlugin` and/or
		:class:`TemplatePlugin` -- the paths provided by ``get_asset_folder`` and ``get_template_folder`` respectively.
		"""
		import os

		if isinstance(self, AssetPlugin):
			static_folder = self.get_asset_folder()
		else:
			static_folder = os.path.join(self._basefolder, "static")

		if isinstance(self, TemplatePlugin):
			template_folder = self.get_template_folder()
		else:
			template_folder = os.path.join(self._basefolder, "templates")

		return dict(
			static_folder=static_folder,
			template_folder=template_folder
		)

	# noinspection PyMethodMayBeStatic
	def is_blueprint_protected(self):
		"""
		Whether a login session by a registered user is needed to access the blueprint's endpoints. Requiring
		a session is the default. Note that this only restricts access to the blueprint's dynamic methods, static files
		are always accessible.

		If you want your blueprint's endpoints to have specific permissions, return ``False`` for this and do your
		permissions checks explicitly.
		"""

		return True


class SettingsPlugin(OctoPrintPlugin):
	"""
	Including the ``SettingsPlugin`` mixin allows plugins to store and retrieve their own settings within OctoPrint's
	configuration.

	Plugins including the mixing will get injected an additional property ``self._settings`` which is an instance of
	:class:`PluginSettingsManager` already properly initialized for use by the plugin. In order for the manager to
	know about the available settings structure and default values upon initialization, implementing plugins will need
	to provide a dictionary with the plugin's default settings through overriding the method :func:`get_settings_defaults`.
	The defined structure will then be available to access through the settings manager available as ``self._settings``.

	If your plugin needs to react to the change of specific configuration values on the fly, e.g. to adjust the log level
	of a logger when the user changes a corresponding flag via the settings dialog, you can override the
	:func:`on_settings_save` method and wrap the call to the implementation from the parent class with retrieval of the
	old and the new value and react accordingly.

	Example:

	.. code-block:: python

	   import octoprint.plugin

	   class MySettingsPlugin(octoprint.plugin.SettingsPlugin, octoprint.plugin.StartupPlugin):
	       def get_settings_defaults(self):
	           return dict(
	               some_setting="foo",
	               some_value=23,
	               sub=dict(
	                   some_flag=True
	               )
	           )

	       def on_settings_save(self, data):
	           old_flag = self._settings.get_boolean(["sub", "some_flag"])

	           octoprint.plugin.SettingsPlugin.on_settings_save(self, data)

	           new_flag = self._settings.get_boolean(["sub", "some_flag"])
	           if old_flag != new_flag:
	               self._logger.info("sub.some_flag changed from {old_flag} to {new_flag}".format(**locals()))

	       def on_after_startup(self):
	           some_setting = self._settings.get(["some_setting"])
	           some_value = self._settings.get_int(["some_value"])
	           some_flag = self._settings.get_boolean(["sub", "some_flag"])
	           self._logger.info("some_setting = {some_setting}, some_value = {some_value}, sub.some_flag = {some_flag}".format(**locals())

	   __plugin_implementation__ = MySettingsPlugin()

	Of course, you are always free to completely override both :func:`on_settings_load` and :func:`on_settings_save` if the
	default implementations do not fit your requirements.


	.. warning::

	   Make sure to protect sensitive information stored by your plugin that only logged in administrators (or users)
	   should have access to via :meth:`~octoprint.plugin.SettingsPlugin.get_settings_restricted_paths`. OctoPrint will
	   return its settings on the REST API even to anonymous clients, but will filter out fields it knows are restricted,
	   therefore you **must** make sure that you specify sensitive information accordingly to limit access as required!
	"""

	config_version_key = "_config_version"
	"""Key of the field in the settings that holds the configuration format version."""

	# noinspection PyMissingConstructor
	def __init__(self):
		self._settings = None
		"""
		The :class:`~octoprint.plugin.PluginSettings` instance to use for accessing the plugin's settings. Injected by
		the plugin core system upon initialization of the implementation.
		"""

	def on_settings_load(self):
		"""
		Loads the settings for the plugin, called by the Settings API view in order to retrieve all settings from
		all plugins. Override this if you want to inject additional settings properties that are not stored within
		OctoPrint's configuration.

		.. note::

		   The default implementation will return your plugin's settings as is, so just in the structure and in the types
		   that are currently stored in OctoPrint's configuration.

		   If you need more granular control here, e.g. over the used data types, you'll need to override this method
		   and iterate yourself over all your settings, using the proper retriever methods on the settings manager
		   to retrieve the data in the correct format.

		   The default implementation will also replace any paths that have been restricted by your plugin through
		   :func:`~octoprint.plugin.SettingsPlugin.get_settings_restricted_paths` with either the provided
		   default value (if one was provided), an empty dictionary (as fallback for restricted dictionaries), an
		   empty list (as fallback for restricted lists) or ``None`` values where necessary.
		   Make sure to do your own restriction if you decide to fully overload this method.

		:return: the current settings of the plugin, as a dictionary
		"""
		from flask_login import current_user
		import copy

		from octoprint.access.permissions import Permissions

		data = copy.deepcopy(self._settings.get_all_data(merged=True))
		if self.config_version_key in data:
			del data[self.config_version_key]

		restricted_paths = self.get_settings_restricted_paths()

		# noinspection PyShadowingNames
		def restrict_path_unless(data, path, condition):
			if not path:
				return

			if condition():
				return

			node = data

			if len(path) > 1:
				for entry in path[:-1]:
					if not entry in node:
						return
					node = node[entry]

			key = path[-1]
			default_value_available = False
			default_value = None
			if isinstance(key, (list, tuple)):
				# key, default_value tuple
				key, default_value = key
				default_value_available = True

			if key in node:
				if default_value_available:
					if callable(default_value):
						default_value = default_value()
					node[key] = default_value
				else:
					if isinstance(node[key], dict):
						node[key] = dict()
					elif isinstance(node[key], (list, tuple)):
						node[key] = []
					else:
						node[key] = None

		conditions = dict(user=lambda: current_user is not None and not current_user.is_anonymous,
		                  admin=lambda: current_user is not None and current_user.has_permission(Permissions.SETTINGS),
		                  never=lambda: False)

		for level, condition in conditions.items():
			paths_for_level = restricted_paths.get(level, [])
			for path in paths_for_level:
				restrict_path_unless(data, path, condition)

		return data

	def on_settings_save(self, data):
		"""
		Saves the settings for the plugin, called by the Settings API view in order to persist all settings
		from all plugins. Override this if you need to directly react to settings changes or want to extract
		additional settings properties that are not stored within OctoPrint's configuration.

		.. note::

		   The default implementation will persist your plugin's settings as is, so just in the structure and in the
		   types that were received by the Settings API view. Values identical to the default settings values
		   will *not* be persisted.

		   If you need more granular control here, e.g. over the used data types, you'll need to override this method
		   and iterate yourself over all your settings, retrieving them (if set) from the supplied received ``data``
		   and using the proper setter methods on the settings manager to persist the data in the correct format.

		Arguments:
		    data (dict): The settings dictionary to be saved for the plugin

		Returns:
		    dict: The settings that differed from the defaults and were actually saved.
		"""
		import octoprint.util

		# get the current data
		current = self._settings.get_all_data()
		if current is None:
			current = dict()

		# merge our new data on top of it
		new_current = octoprint.util.dict_merge(current, data)
		if self.config_version_key in new_current:
			del new_current[self.config_version_key]

		# determine diff dict that contains minimal set of changes against the
		# default settings - we only want to persist that, not everything
		diff = octoprint.util.dict_minimal_mergediff(self.get_settings_defaults(), new_current)

		version = self.get_settings_version()

		to_persist = dict(diff)
		if version:
			to_persist[self.config_version_key] = version

		if to_persist:
			self._settings.set([], to_persist)
		else:
			self._settings.clean_all_data()

		return diff

	# noinspection PyMethodMayBeStatic
	def get_settings_defaults(self):
		"""
		Retrieves the plugin's default settings with which the plugin's settings manager will be initialized.

		Override this in your plugin's implementation and return a dictionary defining your settings data structure
		with included default values.
		"""
		return dict()

	# noinspection PyMethodMayBeStatic
	def get_settings_restricted_paths(self):
		"""
		Retrieves the list of paths in the plugin's settings which be restricted on the REST API.

		Override this in your plugin's implementation to restrict whether a path should only be returned to users with
		the SETTINGS permission, any logged in users, or never on the REST API.

		Return a ``dict`` with the keys ``admin``, ``user``, ``never`` mapping to a list of paths (as tuples or lists of
		the path elements) for which to restrict access via the REST API accordingly. Paths returned for the ``admin``
		key will only be available on the REST API when access with admin rights, ``user`` will only be available when accessed
		as a logged in user. ``never`` will never be returned on the API.

		Example:

		.. code-block:: python

		   def get_settings_defaults(self):
		       return dict(some=dict(admin_only=dict(path="path", foo="foo"),
		                             user_only=dict(path="path", bar="bar")),
		                   another=dict(admin_only=dict(path="path"),
		                                field="field"),
		                   path=dict(to=dict(never=dict(return="return"))))

		   def get_settings_restricted_paths(self):
		       return dict(admin=[["some", "admin_only", "path"], ["another", "admin_only", "path"],],
		                   user=[["some", "user_only", "path"],],
		                   never=[["path", "to", "never", "return"],])

		   # this will make the plugin return settings on the REST API like this for an anonymous user
		   #
		   #     dict(some=dict(admin_only=dict(path=None, foo="foo"),
		   #                    user_only=dict(path=None, bar="bar")),
		   #          another=dict(admin_only=dict(path=None),
		   #                       field="field"),
		   #          path=dict(to=dict(never=dict(return=None))))
		   #
		   # like this for a logged in user
		   #
		   #     dict(some=dict(admin_only=dict(path=None, foo="foo"),
		   #                    user_only=dict(path="path", bar="bar")),
		   #          another=dict(admin_only=dict(path=None),
		   #                       field="field"),
		   #          path=dict(to=dict(never=dict(return=None))))
		   #
		   # and like this for an admin user
		   #
		   #     dict(some=dict(admin_only=dict(path="path", foo="foo"),
		   #                    user_only=dict(path="path", bar="bar")),
		   #          another=dict(admin_only=dict(path="path"),
		   #                       field="field"),
		   #          path=dict(to=dict(never=dict(return=None))))
		"""
		return dict()

	# noinspection PyMethodMayBeStatic
	def get_settings_preprocessors(self):
		"""
		Retrieves the plugin's preprocessors to use for preprocessing returned or set values prior to returning/setting
		them.

		The preprocessors should be provided as a dictionary mapping the path of the values to preprocess
		(hierarchically) to a transform function which will get the value to transform as only input and should return
		the transformed value.

		Example:

		.. code-block:: python

		   def get_settings_defaults(self):
		       return dict(some_key="Some_Value", some_other_key="Some_Value")

		   def get_settings_preprocessors(self):
		       return dict(some_key=lambda x: x.upper()),        # getter preprocessors
		              dict(some_other_key=lambda x: x.lower())   # setter preprocessors

		   def some_method(self):
		       # getting the value for "some_key" should turn it to upper case
		       assert self._settings.get(["some_key"]) == "SOME_VALUE"

		       # the value for "some_other_key" should be left alone
		       assert self._settings.get(["some_other_key"] = "Some_Value"

		       # setting a value for "some_other_key" should cause the value to first be turned to lower case
		       self._settings.set(["some_other_key"], "SOME_OTHER_VALUE")
		       assert self._settings.get(["some_other_key"]) == "some_other_value"

		Returns:
		    (dict, dict): A tuple consisting of two dictionaries, the first being the plugin's preprocessors for
		        getters, the second the preprocessors for setters
		"""
		return dict(), dict()

	# noinspection PyMethodMayBeStatic
	def get_settings_version(self):
		"""
		Retrieves the settings format version of the plugin.

		Use this to have OctoPrint trigger your migration function if it detects an outdated settings version in
		config.yaml.

		Returns:
		    int or None: an int signifying the current settings format, should be incremented by plugins whenever there
		                 are backwards incompatible changes. Returning None here disables the version tracking for the
		                 plugin's configuration.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def on_settings_migrate(self, target, current):
		"""
		Called by OctoPrint if it detects that the installed version of the plugin necessitates a higher settings version
		than the one currently stored in _config.yaml. Will also be called if the settings data stored in config.yaml
		doesn't have version information, in which case the ``current`` parameter will be None.

		Your plugin's implementation should take care of migrating any data by utilizing self._settings. OctoPrint
		will take care of saving any changes to disk by calling `self._settings.save()` after returning from this method.

		This method will be called before your plugin's :func:`on_settings_initialized` method, with all injections already
		having taken place. You can therefore depend on the configuration having been migrated by the time
		:func:`on_settings_initialized` is called.

		Arguments:
		    target (int): The settings format version the plugin requires, this should always be the same value as
		                  returned by :func:`get_settings_version`.
		    current (int or None): The settings format version as currently stored in config.yaml. May be None if
		                  no version information can be found.
		"""
		pass

	def on_settings_cleanup(self):
		"""
		Called after migration and initialization but before call to :func:`on_settings_initialized`.

		Plugins may overwrite this method to perform additional clean up tasks.

		The default implementation just minimizes the data persisted on disk to only contain
		the differences to the defaults (in case the current data was persisted with an older
		version of OctoPrint that still duplicated default data).
		"""
		import octoprint.util
		from octoprint.settings import NoSuchSettingsPath

		try:
			# let's fetch the current persisted config (so only the data on disk,
			# without the defaults)
			config = self._settings.get_all_data(merged=False, incl_defaults=False, error_on_path=True)
		except NoSuchSettingsPath:
			# no config persisted, nothing to do => get out of here
			return

		if config is None:
			# config is set to None, that doesn't make sense, kill it and leave
			self._settings.clean_all_data()
			return

		if self.config_version_key in config and config[self.config_version_key] is None:
			# delete None entries for config version - it's the default, no need
			del config[self.config_version_key]

		# calculate a minimal diff between the settings and the current config -
		# anything already in the settings will be removed from the persisted
		# config, no need to duplicate it
		defaults = self.get_settings_defaults()
		diff = octoprint.util.dict_minimal_mergediff(defaults, config)

		if not diff:
			# no diff to defaults, no need to have anything persisted
			self._settings.clean_all_data()
		else:
			# diff => persist only that
			self._settings.set([], diff)

	def on_settings_initialized(self):
		"""
		Called after the settings have been initialized and - if necessary - also been migrated through a call to
		func:`on_settings_migrate`.

		This method will always be called after the `initialize` method.
		"""
		pass


class EventHandlerPlugin(OctoPrintPlugin):
	"""
	The ``EventHandlerPlugin`` mixin allows OctoPrint plugins to react to any of :ref:`OctoPrint's events <sec-events>`.
	OctoPrint will call the :func:`on_event` method for any event fired on its internal event bus, supplying the
	event type and the associated payload. Please note that until your plugin returns from that method, further event
	processing within OctoPrint will block - the event queue itself is run asynchronously from the rest of OctoPrint,
	but the processing of the events within the queue itself happens consecutively.

	This mixin is especially interesting for plugins which want to react on things like print jobs finishing, timelapse
	videos rendering etc.
	"""

	# noinspection PyMethodMayBeStatic
	def on_event(self, event, payload):
		"""
		Called by OctoPrint upon processing of a fired event on the platform.

		.. warning::

		   Do not perform long-running or even blocking operations in your implementation or you **will** block and break the server.

		Arguments:
		    event (str): The type of event that got fired, see :ref:`the list of events <sec-events-available_events>`
		        for possible values
		    payload (dict): The payload as provided with the event
		"""
		pass


class SlicerPlugin(OctoPrintPlugin):
	"""
	Via the ``SlicerPlugin`` mixin plugins can add support for slicing engines to be used by OctoPrint.

	"""

	# noinspection PyMethodMayBeStatic
	def is_slicer_configured(self):
		"""
		Unless the return value of this method is ``True``, OctoPrint will not register the slicer within the slicing
		sub system upon startup. Plugins may use this to do some start up checks to verify that e.g. the path to
		a slicing binary as set and the binary is executable, or credentials of a cloud slicing platform are properly
		entered etc.
		"""
		return False

	# noinspection PyMethodMayBeStatic
	def get_slicer_properties(self):
		"""
		Plugins should override this method to return a ``dict`` containing a bunch of meta data about the implemented slicer.

		The expected keys in the returned ``dict`` have the following meaning:

		type
		    The type identifier to use for the slicer. This should be a short unique lower case string which will be
		    used to store slicer profiles under or refer to the slicer programmatically or from the API.
		name
		    The human readable name of the slicer. This will be displayed to the user during slicer selection.
		same_device
		    True if the slicer runs on the same device as OctoPrint, False otherwise. Slicers running on the same
		    device will not be allowed to slice on systems with less than two CPU cores (or an unknown number) while a
		    print is running due to performance reasons. Slice requests against slicers running on the same device and
		    less than two cores will result in an error.
		progress_report
		    ``True`` if the slicer can report back slicing progress to OctoPrint ``False`` otherwise.
		source_file_types
		    A list of file types this slicer supports as valid origin file types. These are file types as found in the
		    paths within the extension tree. Plugins may add additional file types through the :ref:`sec-plugins-hook-filemanager-extensiontree` hook.
		    The system will test source files contains in incoming slicing requests via :meth:`octoprint.filemanager.valid_file_type` against the
		    targeted slicer's ``source_file_types``.
		destination_extension
		    The possible extensions of slicing result files.

		Returns:
		    dict: A dict describing the slicer as outlined above.
		"""
		return dict(
			type=None,
			name=None,
			same_device=True,
			progress_report=False,
			source_file_types=["model"],
			destination_extensions=["gco", "gcode", "g"]
		)

	# noinspection PyMethodMayBeStatic
	def get_slicer_extension_tree(self):
		"""
		Fetch additional entries to put into the extension tree for accepted files

		By default, a subtree for ``model`` files with ``stl`` extension is returned. Slicers who want to support
		additional/other file types will want to override this.

		For the extension tree format, take a look at the docs of the :ref:`octoprint.filemanager.extension_tree hook <sec-plugins-hook-filemanager-extensiontree>`.

		Returns: (dict) a dictionary containing a valid extension subtree.
		"""
		from octoprint.filemanager import ContentTypeMapping
		return dict(model=dict(stl=ContentTypeMapping(["stl"], "application/sla")))

	def get_slicer_profiles(self, profile_path):
		"""
		Fetch all :class:`~octoprint.slicing.SlicingProfile` stored for this slicer.

		For compatibility reasons with existing slicing plugins this method defaults to returning profiles parsed from
		.profile files in the plugin's ``profile_path``, utilizing the :func:`SlicingPlugin.get_slicer_profile` method
		of the plugin implementation.

		Arguments:
		    profile_path (str): The base folder where OctoPrint stores this slicer plugin's profiles
		"""

		try:
			from os import scandir
		except ImportError:
			from scandir import scandir

		import octoprint.util

		profiles = dict()
		for entry in scandir(profile_path):
			if not entry.name.endswith(".profile") or octoprint.util.is_hidden_path(entry.name):
				# we are only interested in profiles and no hidden files
				continue

			profile_name = entry.name[:-len(".profile")]
			profiles[profile_name] = self.get_slicer_profile(entry.path)
		return profiles

	# noinspection PyMethodMayBeStatic
	def get_slicer_profiles_lastmodified(self, profile_path):
		import os
		try:
			from os import scandir
		except ImportError:
			from scandir import scandir

		lms = [os.stat(profile_path).st_mtime]
		lms += [os.stat(entry.path).st_mtime for entry in scandir(profile_path) if entry.name.endswith(".profile")]
		return max(lms)

	# noinspection PyMethodMayBeStatic
	def get_slicer_default_profile(self):
		"""
		Should return a :class:`~octoprint.slicing.SlicingProfile` containing the default slicing profile to use with
		this slicer if no other profile has been selected.

		Returns:
		    SlicingProfile: The :class:`~octoprint.slicing.SlicingProfile` containing the default slicing profile for
		        this slicer.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def get_slicer_profile(self, path):
		"""
		Should return a :class:`~octoprint.slicing.SlicingProfile` parsed from the slicing profile stored at the
		indicated ``path``.

		Arguments:
		    path (str): The absolute path from which to read the slicing profile.

		Returns:
		    SlicingProfile: The specified slicing profile.
		"""
		return None

	# noinspection PyMethodMayBeStatic
	def save_slicer_profile(self, path, profile, allow_overwrite=True, overrides=None):
		"""
		Should save the provided :class:`~octoprint.slicing.SlicingProfile` to the indicated ``path``, after applying
		any supplied ``overrides``. If a profile is already saved under the indicated path and ``allow_overwrite`` is
		set to False (defaults to True), an :class:`IOError` should be raised.

		Arguments:
		    path (str): The absolute path to which to save the profile.
		    profile (SlicingProfile): The profile to save.
		    allow_overwrite (boolean): Whether to allow to overwrite an existing profile at the indicated path (True,
		        default) or not (False). If a profile already exists on the path and this is False an
		        :class:`IOError` should be raised.
		    overrides (dict): Profile overrides to apply to the ``profile`` before saving it
		"""
		pass

	# noinspection PyMethodMayBeStatic
	def do_slice(self, model_path, printer_profile, machinecode_path=None, profile_path=None, position=None, on_progress=None, on_progress_args=None, on_progress_kwargs=None):
		"""
		Called by OctoPrint to slice ``model_path`` for the indicated ``printer_profile``. If the ``machinecode_path`` is ``None``,
		slicer implementations should generate it from the provided ``model_path``.

		If provided, the ``profile_path`` is guaranteed by OctoPrint to be a serialized slicing profile created through the slicing
		plugin's own :func:`save_slicer_profile` method.

		If provided, ``position`` will be a ``dict`` containing and ``x`` and a ``y`` key, indicating the position
		the center of the model on the print bed should have in the final sliced machine code. If not provided, slicer
		implementations should place the model in the center of the print bed.

		``on_progress`` will be a callback which expects an additional keyword argument ``_progress`` with the current
		slicing progress which - if progress reporting is supported - the slicing plugin should call like the following:

		.. code-block:: python

		   if on_progress is not None:
		       if on_progress_args is None:
		           on_progress_args = ()
		       if on_progress_kwargs is None:
		           on_progress_kwargs = dict()

		       on_progress_kwargs["_progress"] = your_plugins_slicing_progress
		       on_progress(*on_progress_args, **on_progress_kwargs)

		Please note that both ``on_progress_args`` and ``on_progress_kwargs`` as supplied by OctoPrint might be ``None``,
		so always make sure to initialize those values to sane defaults like depicted above before invoking the callback.

		In order to support external cancellation of an ongoing slicing job via :func:`cancel_slicing`, implementations
		should make sure to track the started jobs via the ``machinecode_path``, if provided.

		The method should return a 2-tuple consisting of a boolean ``flag`` indicating whether the slicing job was
		finished successfully (True) or not (False) and a ``result`` depending on the success of the slicing job.

		For jobs that finished successfully, ``result`` should be a :class:`dict` containing additional information
		about the slicing job under the following keys:

		analysis
		    Analysis result of the generated machine code as returned by the slicer itself. This should match the
		    data structure described for the analysis queue of the matching machine code format, e.g.
		    :class:`~octoprint.filemanager.analysis.GcodeAnalysisQueue` for GCODE files.

		For jobs that did not finish successfully (but not due to being cancelled!), ``result`` should be a :class:`str`
		containing a human readable reason for the error.

		If the job gets cancelled, a :class:`~octoprint.slicing.SlicingCancelled` exception should be raised.

		Returns:
		    tuple: A 2-tuple (boolean, object) as outlined above.

		Raises:
		    SlicingCancelled: The slicing job was cancelled (via :meth:`cancel_slicing`).
		"""
		pass

	# noinspection PyMethodMayBeStatic
	def cancel_slicing(self, machinecode_path):
		"""
		Cancels the slicing to the indicated file.

		Arguments:
		    machinecode_path (str): The absolute path to the machine code file to which to stop slicing to.
		"""
		pass


class ProgressPlugin(OctoPrintPlugin):
	"""
	Via the ``ProgressPlugin`` mixing plugins can let themselves be called upon progress in print jobs or slicing jobs,
	limited to minimally 1% steps.
	"""

	# noinspection PyMethodMayBeStatic
	def on_print_progress(self, storage, path, progress):
		"""
		Called by OctoPrint on minimally 1% increments during a running print job.

		:param string storage:  Location of the file
		:param string path:     Path of the file
		:param int progress:    Current progress as a value between 0 and 100
		"""
		pass

	# noinspection PyMethodMayBeStatic
	def on_slicing_progress(self, slicer, source_location, source_path, destination_location, destination_path, progress):
		"""
		Called by OctoPrint on minimally 1% increments during a running slicing job.

		:param string slicer:               Key of the slicer reporting the progress
		:param string source_location:      Location of the source file
		:param string source_path:          Path of the source file
		:param string destination_location: Location the destination file
		:param string destination_path:     Path of the destination file
		:param int progress:                Current progress as a value between 0 and 100
		"""
		pass
